#!/usr/bin/python2.7
# -*- coding: utf-8 -*-

# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

# Script copyright (C) Campbell Barton, Bastien Montagne
# Modified by Jonas Hauquier for python 2.7 compat and MakeHuman FBX export


import math
#import log
from collections import namedtuple, OrderedDict
from . import encode_bin, data_types

# "Constants"
FBX_VERSION = 7300
FBX_HEADER_VERSION = 1003
FBX_SCENEINFO_VERSION = 100
FBX_TEMPLATES_VERSION = 100

FBX_MODELS_VERSION = 232

FBX_GEOMETRY_VERSION = 124
# Revert back normals to 101 (simple 3D values) for now, 102 (4D + weights) seems not well supported by most apps
# currently, apart from some AD products.
FBX_GEOMETRY_NORMAL_VERSION = 101
FBX_GEOMETRY_BINORMAL_VERSION = 101
FBX_GEOMETRY_TANGENT_VERSION = 101
FBX_GEOMETRY_SMOOTHING_VERSION = 102
FBX_GEOMETRY_VCOLOR_VERSION = 101
FBX_GEOMETRY_UV_VERSION = 101
FBX_GEOMETRY_MATERIAL_VERSION = 101
FBX_GEOMETRY_LAYER_VERSION = 100
FBX_GEOMETRY_SHAPE_VERSION = 100
FBX_DEFORMER_SHAPE_VERSION = 100
FBX_DEFORMER_SHAPECHANNEL_VERSION = 100
FBX_POSE_BIND_VERSION = 100
FBX_DEFORMER_SKIN_VERSION = 101
FBX_DEFORMER_CLUSTER_VERSION = 100
FBX_MATERIAL_VERSION = 102
FBX_TEXTURE_VERSION = 202
FBX_ANIM_KEY_VERSION = 4008

FBX_NAME_CLASS_SEP = b"\x00\x01"
FBX_ANIM_PROPSGROUP_NAME = "d"

FBX_KTIME = 46186158000  # This is the number of "ktimes" in one second (yep, precision over the nanosecond...)


# MAT_CONVERT_BONE = Matrix.Rotation(math.pi / 2.0, 4, 'Z')  # Blender is +Y, FBX is -X.


BLENDER_OBJECT_TYPES_MESHLIKE = {'MESH'}


def getMeshOrientation(config):
    if config.yUpFaceZ:
        return ('Y', 'Z')
    if config.yUpFaceX:
        return ('Y', 'X')
    if config.zUpFaceNegY:
        return ('Z', '-Y')
    if config.zUpFaceX:
        return ('Z', 'X')
    return ('Y', 'Z')

RIGHT_HAND_AXES = {
    # Up, Front -> FBX values (tuples of (axis, sign), Up, Front, Coord).
    ('X',  'Y'):  ((0, 1),  (1, 1),  (2, 1)),
    ('X',  '-Y'): ((0, 1),  (1, -1), (2, -1)),
    ('X',  'Z'):  ((0, 1),  (2, 1),  (1, -1)),
    ('X',  '-Z'): ((0, 1),  (2, -1), (1, 1)),
    ('-X', 'Y'):  ((0, -1), (1, 1),  (2, -1)),
    ('-X', '-Y'): ((0, -1), (1, -1), (2, 1)),
    ('-X', 'Z'):  ((0, -1), (2, 1),  (1, 1)),
    ('-X', '-Z'): ((0, -1), (2, -1), (1, -1)),
    ('Y',  'X'):  ((1, 1),  (0, 1),  (2, -1)),
    ('Y',  '-X'): ((1, 1),  (0, -1), (2, 1)),
    ('Y',  'Z'):  ((1, 1),  (2, 1),  (0, 1)),
    ('Y',  '-Z'): ((1, 1),  (2, -1), (0, -1)),
    ('-Y', 'X'):  ((1, -1), (0, 1),  (2, 1)),
    ('-Y', '-X'): ((1, -1), (0, -1), (2, -1)),
    ('-Y', 'Z'):  ((1, -1), (2, 1),  (0, -1)),
    ('-Y', '-Z'): ((1, -1), (2, -1), (0, 1)),
    ('Z',  'X'):  ((2, 1),  (0, 1),  (1, 1)),
    ('Z',  '-X'): ((2, 1),  (0, -1), (1, -1)),
    ('Z',  'Y'):  ((2, 1),  (1, 1),  (0, -1)),
    ('Z',  '-Y'): ((2, 1),  (1, -1), (0, 1)),  # Blender system!
    ('-Z', 'X'):  ((2, -1), (0, 1),  (1, -1)),
    ('-Z', '-X'): ((2, -1), (0, -1), (1, 1)),
    ('-Z', 'Y'):  ((2, -1), (1, 1),  (0, 1)),
    ('-Z', '-Y'): ((2, -1), (1, -1), (0, -1)),
}


FBX_FRAMERATES = (
    (-1.0, 14),  # Custom framerate.
    (120.0, 1),
    (100.0, 2),
    (60.0, 3),
    (50.0, 4),
    (48.0, 5),
    (30.0, 6),  # BW NTSC.
    (30.0 / 1.001, 9),  # Color NTSC.
    (25.0, 10),
    (24.0, 11),
    (24.0 / 1.001, 13),
    (96.0, 15),
    (72.0, 16),
    (60.0 / 1.001, 17),
)


# ##### Misc utilities #####

# Note: this could be in a utility (math.units e.g.)...

UNITS = {
    "meter": 1.0,  # Ref unit!
    "kilometer": 0.001,
    "millimeter": 1000.0,
    "foot": 1.0 / 0.3048,
    "inch": 1.0 / 0.0254,
    "turn": 1.0,  # Ref unit!
    "degree": 360.0,
    "radian": math.pi * 2.0,
    "second": 1.0,  # Ref unit!
    "ktime": FBX_KTIME,
}


def units_convertor(u_from, u_to):
    """Return a convertor between specified units."""
    conv = UNITS[u_to] / UNITS[u_from]
    return lambda v: v * conv


def units_convertor_iter(u_from, u_to):
    """Return an iterable convertor between specified units."""
    conv = units_convertor(u_from, u_to)

    def convertor(it):
        for v in it:
            yield(conv(v))

    return convertor


def matrix4_to_array(mat):
    """Concatenate matrix's columns into a single, flat tuple"""
    # blender matrix is row major, fbx is col major so transpose on write
    return tuple(f for v in mat.transposed() for f in v)


def array_to_matrix4(arr):
    """Convert a single 16-len tuple into a valid 4D Blender matrix"""
    # Blender matrix is row major, fbx is col major so transpose on read
    return Matrix(tuple(zip(*[iter(arr)]*4))).transposed()


def similar_values(v1, v2, e=1e-6):
    """Return True if v1 and v2 are nearly the same."""
    if v1 == v2:
        return True
    return ((abs(v1 - v2) / max(abs(v1), abs(v2))) <= e)


def similar_values_iter(v1, v2, e=1e-6):
    """Return True if iterables v1 and v2 are nearly the same."""
    if v1 == v2:
        return True
    for v1, v2 in zip(v1, v2):
        if (v1 != v2) and ((abs(v1 - v2) / max(abs(v1), abs(v2))) > e):
            return False
    return True

def vcos_transformed_gen(raw_cos, m=None):
    # Note: we could most likely get much better performances with numpy, but will leave this as TODO for now.
    gen = zip(*(iter(raw_cos),) * 3)
    return gen if m is None else (m * Vector(v) for v in gen)

def nors_transformed_gen(raw_nors, m=None):
    # Great, now normals are also expected 4D!
    # XXX Back to 3D normals for now!
    # gen = zip(*(iter(raw_nors),) * 3 + (_infinite_gen(1.0),))
    gen = zip(*(iter(raw_nors),) * 3)
    return gen if m is None else (m * Vector(v) for v in gen)


'''
# ##### UIDs code. #####

# ID class (mere int).
class UUID(int):
    pass


# UIDs storage.
_keys_to_uuids = {}
_uuids_to_keys = {}


def _key_to_uuid(uuids, key):
    # TODO: Check this is robust enough for our needs!
    # Note: We assume we have already checked the related key wasn't yet in _keys_to_uids!
    #       As int64 is signed in FBX, we keep uids below 2**63...
    if isinstance(key, int) and 0 <= key < 2**63:
        # We can use value directly as id!
        uuid = key
    else:
        uuid = hash(key)
        if uuid < 0:
            uuid = -uuid
        if uuid >= 2**63:
            uuid //= 2
    # Try to make our uid shorter!
    if uuid > int(1e9):
        t_uuid = uuid % int(1e9)
        if t_uuid not in uuids:
            uuid = t_uuid
    # Make sure our uuid *is* unique.
    if uuid in uuids:
        inc = 1 if uuid < 2**62 else -1
        while uuid in uuids:
            uuid += inc
            if 0 > uuid >= 2**63:
                # Note that this is more that unlikely, but does not harm anyway...
                raise ValueError("Unable to generate an UUID for key {}".format(key))
    return UUID(uuid)


def get_fbx_uuid_from_key(key):
    """
    Return an UUID for given key, which is assumed hasable.
    """
    uuid = _keys_to_uuids.get(key, None)
    if uuid is None:
        uuid = _key_to_uuid(_uuids_to_keys, key)
        _keys_to_uuids[key] = uuid
        _uuids_to_keys[uuid] = key
    return uuid


# XXX Not sure we'll actually need this one?
def get_key_from_fbx_uuid(uuid):
    """
    Return the key which generated this uid.
    """
    assert(uuid.__class__ == UUID)
    return _uuids_to_keys.get(uuid, None)


    """Return (stack/layer, ID, fbxprop, item) curve key."""
    return "|".join((get_blender_anim_id_base(scene, ref_id), obj_key, fbx_prop_name,
                     fbx_prop_item_name, "AnimCurve"))
'''

# ##### Element generators. #####

# Note: elem may be None, in this case the element is not added to any parent.
def elem_empty(elem, name):
    sub_elem = encode_bin.FBXElem(name)
    if elem is not None:
        elem.elems.append(sub_elem)
    return sub_elem


def _elem_data_single(elem, name, value, func_name):

    sub_elem = elem_empty(elem, name)

    from . import fbx_utils
    fbx_utils.debugWrite(name, "elem_data_single")
    fbx_utils.debugWrite(value, "elem_data_single")

    getattr(sub_elem, func_name)(value)
    return sub_elem


def _elem_data_vec(elem, name, value, func_name):

    from . import fbx_utils
    fbx_utils.debugWrite(name, "elem_data_single")
    fbx_utils.debugWrite(value, "elem_data_single")

    sub_elem = elem_empty(elem, name)
    func = getattr(sub_elem, func_name)
    for v in value:
        func(v)
    return sub_elem

def get_child_element(parentelem, name):
    assert(isinstance(name, bytes))
    for child in parentelem.elems:
        if child.id == name:
            return child
    return None

def elem_data_single_bool(elem, name, value):
    return _elem_data_single(elem, name, value, "add_bool")


def elem_data_single_int16(elem, name, value):
    return _elem_data_single(elem, name, value, "add_int16")


def elem_data_single_int32(elem, name, value):
    return _elem_data_single(elem, name, value, "add_int32")


def elem_data_single_int64(elem, name, value):
    return _elem_data_single(elem, name, value, "add_int64")


def elem_data_single_float32(elem, name, value):
    return _elem_data_single(elem, name, value, "add_float32")


def elem_data_single_float64(elem, name, value):
    return _elem_data_single(elem, name, value, "add_float64")


def elem_data_single_bytes(elem, name, value):
    return _elem_data_single(elem, name, value, "add_bytes")


def elem_data_single_string(elem, name, value):
    return _elem_data_single(elem, name, value, "add_string")


def elem_data_single_string_unicode(elem, name, value):
    return _elem_data_single(elem, name, value, "add_string_unicode")


def elem_data_single_bool_array(elem, name, value):
    return _elem_data_single(elem, name, value, "add_bool_array")


def elem_data_single_int32_array(elem, name, value):
    return _elem_data_single(elem, name, value, "add_int32_array")


def elem_data_single_int64_array(elem, name, value):
    return _elem_data_single(elem, name, value, "add_int64_array")


def elem_data_single_float32_array(elem, name, value):
    return _elem_data_single(elem, name, value, "add_float32_array")


def elem_data_single_float64_array(elem, name, value):
    return _elem_data_single(elem, name, value, "add_float64_array")


def elem_data_single_byte_array(elem, name, value):
    return _elem_data_single(elem, name, value, "add_byte_array")


def elem_data_vec_float64(elem, name, value):
    return _elem_data_vec(elem, name, value, "add_float64")


# ##### Generators for standard FBXProperties70 properties. #####

def elem_properties(elem):
    return elem_empty(elem, b"Properties70")


# Properties definitions, format: (b"type_1", b"label(???)", "name_set_value_1", "name_set_value_2", ...)
# XXX Looks like there can be various variations of formats here... Will have to be checked ultimately!
#     Also, those "custom" types like 'FieldOfView' or 'Lcl Translation' are pure nonsense,
#     these are just Vector3D ultimately... *sigh* (again).
FBX_PROPERTIES_DEFINITIONS = {
    # Generic types.
    "p_bool": (b"bool", b"", "add_int32"),  # Yes, int32 for a bool (and they do have a core bool type)!!!
    "p_integer": (b"int", b"Integer", "add_int32"),
    "p_ulonglong": (b"ULongLong", b"", "add_int64"),
    "p_double": (b"double", b"Number", "add_float64"),  # Non-animatable?
    "p_number": (b"Number", b"", "add_float64"),  # Animatable-only?
    "p_enum": (b"enum", b"", "add_int32"),
    "p_vector_3d": (b"Vector3D", b"Vector", "add_float64", "add_float64", "add_float64"),  # Non-animatable?
    "p_vector": (b"Vector", b"", "add_float64", "add_float64", "add_float64"),  # Animatable-only?
    "p_color_rgb": (b"ColorRGB", b"Color", "add_float64", "add_float64", "add_float64"),  # Non-animatable?
    "p_color": (b"Color", b"", "add_float64", "add_float64", "add_float64"),  # Animatable-only?
    "p_string": (b"KString", b"", "add_string_unicode"),
    "p_string_url": (b"KString", b"Url", "add_string_unicode"),
    "p_string_xrefurl": (b"KString", b"XrefUrl", "add_string_unicode"),
    "p_timestamp": (b"KTime", b"Time", "add_int64"),
    "p_datetime": (b"DateTime", b"", "add_string_unicode"),
    # Special types.
    "p_object": (b"object", b""),  # XXX Check this! No value for this prop??? Would really like to know how it works!
    "p_compound": (b"Compound", b""),
    # Specific types (sic).
    # ## Objects (Models).
    "p_lcl_translation": (b"Lcl Translation", b"", "add_float64", "add_float64", "add_float64"),
    "p_lcl_rotation": (b"Lcl Rotation", b"", "add_float64", "add_float64", "add_float64"),
    "p_lcl_scaling": (b"Lcl Scaling", b"", "add_float64", "add_float64", "add_float64"),
    "p_visibility": (b"Visibility", b"", "add_float64"),
    "p_visibility_inheritance": (b"Visibility Inheritance", b"", "add_int32"),
    # ## Cameras!!!
    "p_roll": (b"Roll", b"", "add_float64"),
    "p_opticalcenterx": (b"OpticalCenterX", b"", "add_float64"),
    "p_opticalcentery": (b"OpticalCenterY", b"", "add_float64"),
    "p_fov": (b"FieldOfView", b"", "add_float64"),
    "p_fov_x": (b"FieldOfViewX", b"", "add_float64"),
    "p_fov_y": (b"FieldOfViewY", b"", "add_float64"),
}

def get_ascii_properties(properties, indent=0):
    result = []
    for p in properties:
        if len(p) < 3:
            continue
        if len(p) == 3:
            name, ptype, value = p
            animatable = False
            custom = False
        elif len(p) == 4:
            name, ptype, value, animatable = p
            custom = False
        else:
            name, ptype, value, animatable, custom = p

        result.append( indent*'    ' + get_ascii_property(name, ptype, value, animatable, custom) )

    return '\n'.join(result)

def get_ascii_property(name, ptype, value, animatable=False, custom=False):
    if isinstance(value, tuple):
        value = list(value)
    elif not isinstance(value, list):
        value = [value]

    flags = _elem_props_flags(animatable, custom)
    ptype = FBX_PROPERTIES_DEFINITIONS[ptype]

    if len(ptype) > 2 and 'string' in ptype[2]:
        value = ['"%s"' % v for v in value]
    elif len(ptype) > 2 and 'int' in ptype[2]:
        value = [int(v) for v in value]

    return 'P: "%s", "%s", "%s", "%s", %s' % (name, ptype[0], ptype[1], flags, (','.join([str(v) for v in value])))

def _elem_props_set(elem, ptype, name, value, flags):
    p = elem_data_single_string(elem, b"P", name)
    for t in ptype[:2]:
        p.add_string(t)
    p.add_string(flags)
    if len(ptype) == 3:
        getattr(p, ptype[2])(value)
    elif len(ptype) > 3:
        # We assume value is iterable, else it's a bug!
        for callback, val in zip(ptype[2:], value):
            #log.debug(p)
            #log.debug(type(p))
            #log.debug(callback)
            #log.debug(type(callback))
            #log.debug(val)
            #log.debug(type(val))
            getattr(p, callback)(val)


def _elem_props_flags(animatable, custom):
    if animatable and custom:
        return b"AU"
    elif animatable:
        return b"A"
    elif custom:
        return b"U"
    return b""


def elem_props_set(elem, ptype, name, value=None, animatable=False, custom=False):
    ptype = FBX_PROPERTIES_DEFINITIONS[ptype]
    #log.debug('propset %s %s %s', name, value, type(value))
    _elem_props_set(elem, ptype, name, value, _elem_props_flags(animatable, custom))


def elem_props_compound(elem, cmpd_name, custom=False):
    def _setter(ptype, name, value, animatable=False, custom=False):
        name = cmpd_name + b"|" + name
        elem_props_set(elem, ptype, name, value, animatable=animatable, custom=custom)

    elem_props_set(elem, "p_compound", cmpd_name, custom=custom)
    return _setter


def elem_props_template_init(templates, template_type):
    """
    Init a writing template of given type, for *one* element's properties.
    """
    ret = OrderedDict()
    tmpl = templates.get(template_type)
    if tmpl is not None:
        written = tmpl.written[0]
        props = tmpl.properties
        ret = OrderedDict((name, [val, ptype, anim, written]) for name, (val, ptype, anim) in props.items())
    return ret


def elem_props_template_set(template, elem, ptype_name, name, value, animatable=False):
    """
    Only add a prop if the same value is not already defined in given template.
    Note it is important to not give iterators as value, here!
    """
    ptype = FBX_PROPERTIES_DEFINITIONS[ptype_name]
    if len(ptype) > 3:
        value = tuple(value)
    tmpl_val, tmpl_ptype, tmpl_animatable, tmpl_written = template.get(name, (None, None, False, False))
    # Note animatable flag from template takes precedence over given one, if applicable.
    if tmpl_ptype is not None:
        if (tmpl_written and
            ((len(ptype) == 3 and (tmpl_val, tmpl_ptype) == (value, ptype_name)) or
             (len(ptype) > 3 and (tuple(tmpl_val), tmpl_ptype) == (value, ptype_name)))):
            return  # Already in template and same value.
        _elem_props_set(elem, ptype, name, value, _elem_props_flags(tmpl_animatable, False))
        template[name][3] = True
    else:
        _elem_props_set(elem, ptype, name, value, _elem_props_flags(animatable, False))


def elem_props_template_finalize(template, elem):
    """
    Finalize one element's template/props.
    Issue is, some templates might be "needed" by different types (e.g. NodeAttribute is for lights, cameras, etc.),
    but values for only *one* subtype can be written as template. So we have to be sure we write those for the other
    subtypes in each and every elements, if they are not overriden by that element.
    Yes, hairy, FBX that is to say. When they could easily support several subtypes per template... :(
    """
    for name, (value, ptype_name, animatable, written) in template.items():
        if written:
            continue
        ptype = FBX_PROPERTIES_DEFINITIONS[ptype_name]
        _elem_props_set(elem, ptype, name, value, _elem_props_flags(animatable, False))


# ##### Templates #####
# TODO: check all those "default" values, they should match Blender's default as much as possible, I guess?

FBXTemplate = namedtuple("FBXTemplate", ("type_name", "prop_type_name", "properties", "nbr_users", "written"))

def get_properties(properties):
    for p in properties:
        #log.debug(p)
        if len(p) < 2:
            continue
        if len(p) == 2:
            name = p[0]
            value, ptype, animatable = p[1]
            custom = False
        if len(p) == 3:
            name, ptype, value = p
            animatable = False
            custom = False
        if len(p) == 4:
            name, ptype, value, animatable = p
            custom = False
        if len(p) > 4:
            name, ptype, value, animatable, custom = p

        yield (name, ptype, value, animatable, custom)

def fbx_template_generate(definitionsNode, objectType_name, users_count, propertyTemplate_name=None, properties=[]):
    template = elem_data_single_string(definitionsNode, b"ObjectType", objectType_name)
    elem_data_single_int32(template, b"Count", users_count)

    if propertyTemplate_name and len(properties) > 0:
        elem = elem_data_single_string(template, b"PropertyTemplate", propertyTemplate_name)
        props = elem_properties(elem)

        #log.debug("--- fbx_template_generate ---")
        #log.debug(len(properties))

        for name, ptype, value, animatable, custom in get_properties(properties):
            try:
                elem_props_set(props, ptype, name, value, animatable, custom)
            except Exception as e:
                #log.debug("FBX: Failed to write template prop (%r) (%s)", e, str((props, ptype, name, value, animatable)))
                pass


def fbx_name_class(name, cls=None):
    if cls is None:
        cls,name = name.split(b'::')
    #log.debug(cls)
    #log.debug(type(cls))
    #log.debug(name)
    #log.debug(type(name))
    #log.debug(type(FBX_NAME_CLASS_SEP))
    return FBX_NAME_CLASS_SEP.join((name, cls))




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
**Project Name:**      MakeHuman

**Product Home Page:** http://www.makehumancommunity.org/

**Github Code Home Page:**    https://github.com/makehumancommunity/

**Authors:**           Glynn Clements

**Copyright(c):**      MakeHuman Team 2001-2020

**Licensing:**         AGPL3

    This file is part of MakeHuman (www.makehumancommunity.org).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.


Abstract
--------

TODO
"""

import os
import numpy as np

import gui3d
import mh
import gui
import algos3d
from core import G
import log

class TargetsTree(gui.TreeView):
    def __init__(self):
        super(TargetsTree, self).__init__()
        self.root = self.addTopLevel('targets')

    @staticmethod
    def getItemPath(item):
        path = []
        while item is not None:
            path.append(item.text)
            item = item.parent
        path = mh.getSysDataPath(os.path.join(*reversed(path)))
        return path

    def populate(self, item):
        path = self.getItemPath(item)
        for name in sorted(os.listdir(path)):
            if name[0] == '.':
                continue
            pathname = os.path.join(path, name)
            if os.path.isdir(pathname):
                item.addChild(name, True)
            elif name.lower().endswith('.target') and os.path.isfile(pathname):
                item.addChild(name, False)

class TargetsTaskView(gui3d.TaskView):
    def __init__(self, category):
        super(TargetsTaskView, self).__init__(category, 'Targets')

        self.items = {}

        self.left.child.setSizePolicy(gui.SizePolicy.MinimumExpanding, gui.SizePolicy.MinimumExpanding)

        self.clear = self.addLeftWidget(gui.Button('Clear'))

        self.targets = self.addLeftWidget(TargetsTree())
        self.targets.setHeaderHidden(True)
        self.targets.resizeColumnToContents(0)
        self.targets.setSizePolicy(gui.SizePolicy.Ignored, gui.SizePolicy.Expanding)
        self.left.layout.setStretchFactor(self.targets, 1)

        self.itemsBox = gui.GroupBox('Selected Targets:')
        self.itemsList = gui.TextView('')
        self.itemsList.setWordWrap(False)
        self.itemsList.setSizePolicy(gui.SizePolicy.Ignored, gui.SizePolicy.Preferred)
        self.itemsBox.addWidget(self.itemsList)
        self.addRightWidget(self.itemsBox)

        @self.targets.mhEvent
        def onActivate(item):
            path = self.targets.getItemPath(item)
            log.message('target: %s', path)
            if item.text not in self.items:
                self.items[item.text] = [item, path]
                self.showTarget(path)
            else:
                self.clearColor()
                del self.items[item.text]
                item.setSelected(False)
                for itm in self.items:
                    self.showTarget(self.items[itm][1])
            itemsString = '\n'.join(key.split('.')[0] for key in sorted(self.items))
            for key in self.items:
                self.items[key][0].setSelected(True)
            self.itemsList.setText(itemsString)

        @self.targets.mhEvent
        def onExpand(item):
            self.targets.populate(item)
            self.targets.resizeColumnToContents(0)

        @self.clear.mhEvent
        def onClicked(event):
            self.clearColor()
            self.itemsList.setText('')
            for key in self.items:
                self.items[key][0].setSelected(False)
            self.items.clear()

    def clearColor(self):
        mesh = G.app.selectedHuman.meshData
        mesh.color[...] = (255,255,255,255)
        mesh.markCoords(colr = True)
        mesh.sync_all()

    def showTarget(self, path):
        mesh = G.app.selectedHuman.meshData
        target = algos3d.getTarget(mesh, path)
        sizes = np.sqrt(np.sum(target.data ** 2, axis = -1))
        sizes /= np.amax(sizes)
        val = sizes * 2 - 1
        del sizes
        red = np.maximum(val, 0)
        blue = np.maximum(-val, 0)
        green = 1.0 - red - blue
        del val
        color = np.array([red,green,blue,np.ones_like(red)]).T
        color = (color * 255.99).astype(np.uint8)
        mesh.color[target.verts,:] = color
        mesh.markCoords(target.verts, colr = True)
        mesh.sync_all()

    def onShow(self, event):
        gui3d.TaskView.onShow(self, event)
        gui3d.app.actions.wireframe.setDisabled(True)
        if not gui3d.app.actions.wireframe.isChecked():
            G.app.selectedHuman.setSolid(False)

    def onHide(self, event):
        gui3d.TaskView.onHide(self, event)
        if not gui3d.app.actions.wireframe.isChecked():
            G.app.selectedHuman.setSolid(True)
        gui3d.app.actions.wireframe.setDisabled(False)

def load(app):
    category = app.getCategory('Utilities')
    taskview = category.addTask(TargetsTaskView(category))

def unload(app):
    pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
**Project Name:**      MakeHuman

**Product Home Page:** http://www.makehumancommunity.org/

**Github Code Home Page:**    https://github.com/makehumancommunity/

**Authors:**           Thomas Larsson, Jonas Hauquier

**Copyright(c):**      MakeHuman Team 2001-2020

**Licensing:**         AGPL3

    This file is part of MakeHuman (www.makehumancommunity.org).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.


Abstract
--------

BVH exporter.
Supports exporting of selected skeleton and animations in BVH format.
"""

import bvh

from export import Exporter, ExportConfig
import log
from core import G

import os

# TODO add options such as z-up, feetonground, etc

class BvhConfig(ExportConfig):

    def __init__(self):
        ExportConfig.__init__(self)
        self.useRelPaths = True

class ExporterBVH(Exporter):
    def __init__(self):
        Exporter.__init__(self)
        self.group = "rig"
        self.name = "Biovision Hierarchy BVH"
        self.filter = "Biovision Hierarchy (*.bvh)"
        self.fileExtension = "bvh"

    def build(self, options, taskview):
        import gui
        self.taskview       = taskview
        #self.exportAnimations = options.addWidget(gui.CheckBox("Animations", True))
        self.feetOnGround = options.addWidget(gui.CheckBox("Feet on ground", True))

    def getConfig(self):
        cfg = BvhConfig()
        cfg.feetOnGround      = self.feetOnGround.selected
        cfg.scale,cfg.unit    = self.taskview.getScale()
        #cfg.exportAnimations = self.exportAnimations.selected
        cfg.exportAnimations = False

        return cfg

    def export(self, human, filename):
        if not human.getSkeleton():
            G.app.prompt('Error', 'You did not select a skeleton from the library.', 'OK')
            return

        skel = human.getSkeleton()
        cfg = self.getConfig()
        cfg.setHuman(human)

        if cfg.exportAnimations and len(human.getAnimations()) > 0:
            baseFilename = os.path.splitext(filename("bvh"))[0]
            for animName in human.getAnimations():
                fn = baseFilename + "_%s.bvh" % animName
                log.message("Exporting file %s.", fn)
                bvhData = bvh.createFromSkeleton(skel, human.getAnimation(animName))
                if cfg.scale != 1:
                    bvhData.scale(cfg.scale)
                if cfg.feetOnGround:
                    bvhData.offset(cfg.offset)
                bvhData.writeToFile(fn)
        else:
            fn = filename("bvh")
            log.message("Exporting file %s.", fn)
            if human.isPosed():
                bvhData = bvh.createFromSkeleton(skel, human.getActiveAnimation())
            else:
                bvhData = bvh.createFromSkeleton(skel)
            if cfg.scale != 1:
                bvhData.scale(cfg.scale)
            if cfg.feetOnGround:
                bvhData.offset(cfg.offset)
            bvhData.writeToFile(fn)

def load(app):
    app.addExporter(ExporterBVH())

def unload(app):
    pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
**Project Name:**      MakeHuman

**Product Home Page:** http://www.makehumancommunity.org/

**Github Code Home Page:**    https://github.com/makehumancommunity/

**Authors:**           Glynn Clements

**Copyright(c):**      MakeHuman Team 2001-2020

**Licensing:**         AGPL3

    This file is part of MakeHuman (www.makehumancommunity.org).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.


Abstract
--------

TODO
"""

import gui3d
import mh
import gui
import log
from core import G

class LevelRadioButton(gui.RadioButton):
    def __init__(self, group, level):
        name = log.getLevelName(level).capitalize()
        super(LevelRadioButton, self).__init__(group, name, not group)
        self.level = level

    def onClicked(self, _dummy = None):
        if self.selected:
            G.app.log_window.setLevel(self.level)

class LoggingTaskView(gui3d.TaskView):
    def __init__(self, category):
        super(LoggingTaskView, self).__init__(category, 'Logs')
        self.addTopWidget(G.app.log_window)
        self.groupBox = self.addLeftWidget(gui.GroupBox('Level'))
        group = []
        for level in [log.DEBUG, log.MESSAGE, log.NOTICE, log.WARNING, log.ERROR]:
            radio = self.groupBox.addWidget(LevelRadioButton(group, level))
        self.copy = self.addLeftWidget(gui.Button('Copy to Clipboard'))

        @self.copy.mhEvent
        def onClicked(self, _dummy = None):
            strings = G.app.log_window.getSelectedItems()
            if not strings:
                G.app.status('No log items selected')
                return
            text = ''.join(string + '\n' for string in strings)
            G.app.clipboard().setText(text)

def load(app):
    category = app.getCategory('Utilities')
    taskview = category.addTask(LoggingTaskView(category))

def unload(app):
    pass




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

""" 
**Project Name:**      MakeHuman

**Product Home Page:** http://www.makehumancommunity.org/

**Github Code Home Page:**    https://github.com/makehumancommunity/

**Authors:**           Jonas Hauquier

**Copyright(c):**      MakeHuman Team 2001-2020

**Licensing:**         AGPL3

    This file is part of MakeHuman Community (www.makehumancommunity.org).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.


Abstract
--------

Mesh operations such as calculating volume and surface measures.
"""

import numpy as np
import math

def calculateSurface(mesh, vertGroups=None, faceMask=None):
    """
    Calculate surface area of a mesh. Specify vertGroups or faceMask to
    calculate area of a subset of the mesh and filter out other faces.
    """
    if vertGroups is not None:
        f_idx = mesh.getFacesForGroups(vertGroups)
        fvert = mesh.fvert[f_idx]
    elif faceMask is not None:
        f_idx = np.argwhere(faceMask)[...,0]
        fvert = mesh.fvert[f_idx]
    else:
        fvert = mesh.fvert

    if mesh.vertsPerPrimitive == 4:
        # Split quads in triangles (assumes clockwise ordering of verts)
        t1 = fvert[:,[0,1,2]]
        t2 = fvert[:,[2,3,0]]
        v1 = mesh.coord[t1]
        v2 = mesh.coord[t2]

        l1 = _sideLengthsFromTris(v1)
        l2 = _sideLengthsFromTris(v2)
        l = np.vstack([l1,l2])

        return _surfaceOfTris(l)
    elif mesh.vertsPerPrimitive == 3:
        v = mesh.coord[fvert]
        l = _sideLengthsFromTris(v)
        return _surfaceOfTris(l)
    else:
        raise RuntimeError("Only supports meshes with triangle or quad primitives.")

def calculateVolume(mesh, vertGroups=None, faceMask=None):
    """
    Calculate the volume of a mesh.
    Mesh is expected to be closed.
    """
    if vertGroups is not None:
        f_idx = mesh.getFacesForGroups(vertGroups)
        fvert = mesh.fvert[f_idx]
    elif faceMask is not None:
        f_idx = np.argwhere(faceMask)[...,0]
        fvert = mesh.fvert[f_idx]
    else:
        fvert = mesh.fvert

    if mesh.vertsPerPrimitive == 4:
        # Split quads in triangles (assumes clockwise ordering of verts)
        t1 = fvert[:,[0,1,2]]
        t2 = fvert[:,[2,3,0]]
        v1 = mesh.coord[t1]
        v2 = mesh.coord[t2]

        v = np.vstack([v1,v2])
        return _signedVolumeFromTris(v)
    elif mesh.vertsPerPrimitive == 3:
        v = mesh.coord[fvert]
        return _signedVolumeFromTris(v)
    else:
        raise RuntimeError("Only supports meshes with triangle or quad primitives.")

def _sideLengthsFromTris(triVects):
    """
    Calculate lengths of the sides of triangles specified by their vectors
    in clockwise fashion.
    triVects = [ [T1V1, T1V2, T1V3], [T2V1, T2V2, T2V3], ... ]
    with Ti a triangle, Vi a triange vector, defined in clockwise fashion
    and each vector (TiVi) an array [x, y, z] with vector coordinates

    Returns a list [ [T1L1, T1L2, T1L3], [T2L1, T2L2, T2L3], ...]
    with Ti a triangle (in the same order as in the input), and Li the length of
    side i (a float)
    """
    v = triVects
    s = np.zeros(v.shape, dtype=np.float32)

    # Get side vectors
    s[:,0] = v[:,1] - v[:,0]
    s[:,1] = v[:,2] - v[:,1]
    s[:,2] = v[:,0] - v[:,2]

    # Calculate lengths of sides
    l = s[:,:,0]*s[:,:,0] + s[:,:,1]*s[:,:,1] + s[:,:,2]*s[:,:,2]
    l = np.sqrt(l)

    return l

def _surfaceOfTris(triSideLengths):
    """
    Calculate total surface area of triangles with sides of specified lengths
    triSideLengths should be an array of layout 
    [ [T1L1, T1L2, T1L3], [T2L1, T2L2, T2L3], ... ]
    with Ti a triangle, and Li the length of the ith side of the triangle
    TiLi should be a float.

    Returns a float representing the total surface area.
    """
    l = triSideLengths

    # Heron's formula
    o = ( l[:,0]  +l[:,1]  +l[:,2]) * \
        ( l[:,0]  +l[:,1]  -l[:,2]) * \
        (-l[:,0]  +l[:,1]  +l[:,2]) * \
        ( l[:,0]  -l[:,1]  +l[:,2])
    o = np.sqrt(o)/4

    return np.sum(o)

def _signedVolumeFromTris(triVects):
    """
    Calculate volume of a set of triangles by summing signed volumes of 
    tetrahedrons between those triangles and the origin.
    """
    v = triVects

    v321 = v[:,2,0] * v[:,1,1] * v[:,0,2]
    v231 = v[:,1,0] * v[:,2,1] * v[:,0,2]
    v312 = v[:,2,0] * v[:,0,1] * v[:,1,2]
    v132 = v[:,0,0] * v[:,2,1] * v[:,1,2]
    v213 = v[:,1,0] * v[:,0,1] * v[:,2,2]
    v123 = v[:,0,0] * v[:,1,1] * v[:,2,2]
    signedVolume = -v321 + v231 + v312 - v132 - v213 + v123
    signedVolume /= 6.0

    vol = np.sum(signedVolume)
    return math.fabs(vol)

def findVertIndex(mesh, vert):
    """
    Find the index of specified vertex (as an [x, y, z] array) within mesh.
    """
    matches = list(np.where(mesh.coord == vert)[0])
    return [idx for idx in set(matches) if matches.count(idx) > 2]




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
**Project Name:**      MakeHuman

**Product Home Page:** http://www.makehumancommunity.org/

**Github Code Home Page:**    https://github.com/makehumancommunity/

**Authors:**           Glynn Clements, Jonas Hauquier, Aranuvir

**Copyright(c):**      MakeHuman Team 2001-2020

**Licensing:**         AGPL3

    This file is part of MakeHuman (www.makehumancommunity.org).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.



Abstract
--------

TODO
"""

import sys
import traceback

from PyQt5 import QtCore, QtGui, QtWidgets

import gui
import gui3d
from core import G
import importlib.util
import mh
from language import language

hasIpython = importlib.util.find_spec('qtconsole') is not None and importlib.util.find_spec('IPython') is not None
if hasIpython:
    from . import ipythonconsole


MAX_COMPLETIONS = -1

class ShellTextEdit(gui.TextEdit):
    def tabPressed(self):
        self.callEvent('onTabPressed', self)
        return True

class ShellTaskView(gui3d.TaskView):

    def __init__(self, category):
        super(ShellTaskView, self).__init__(category, 'Shell')
        self.globals = {'G': G}
        self.history = []
        self.histitem = None

        #Register shortcut Shift S for the shell tab
        action = gui.Action('shell', language.getLanguageString('Python Shell'),
                            lambda: mh.changeTask('Utilities', 'Shell'))
        G.app.mainwin.addAction(action)
        mh.setShortcut(mh.Modifiers.SHIFT, mh.Keys.s, action)

        if hasIpython:
            # Use the more advanced Ipython console
            self.console = self.addTopWidget(ipythonconsole.IPythonConsoleWidget())

        else:

            # Fall back to old console
            self.console = None
            self.main = self.addTopWidget(QtWidgets.QWidget())
            self.layout = QtWidgets.QGridLayout(self.main)
            self.layout.setRowStretch(0, 0)
            self.layout.setRowStretch(1, 0)
            self.layout.setColumnStretch(0, 1)
            self.layout.setColumnStretch(1, 0)

            self.text = gui.DocumentEdit()
            self.text.setSizePolicy(
                QtWidgets.QSizePolicy.Expanding,
                QtWidgets.QSizePolicy.Expanding)
            self.layout.addWidget(self.text, 0, 0, 1, 2)

            self.line = ShellTextEdit()
            self.line.setFocusPolicy(QtCore.Qt.StrongFocus)
            self.layout.addWidget(self.line, 1, 0, 1, 1)
            self.globals = {'G': G}

            self.clear = gui.Button("Clear")
            self.layout.addWidget(self.clear, 1, 1, 1, 1)

            @self.line.mhEvent
            def onActivate(text):
                self.execute(text)
                self.history.append(text)
                self.histitem = None
                self.line.setText('')

            @self.line.mhEvent
            def onTabPressed(edit):
                def _longest_common_substring(s1, s2):
                    """
                    This is simply the O(n) left-aligned version
                    """
                    limit = min(len(s1), len(s2))
                    i = 0
                    while i < limit and s1[i] == s2[i]:
                        i += 1
                    return s1[:i]

                def _largest_common(strings):
                    strings = list(strings)
                    try:
                        strings.remove('...')
                    except:
                        pass

                    if len(strings) == 0:
                        return ""
                    elif len(strings) == 1:
                        return strings[0]
                    else:
                        result = strings[0]
                        for s in strings[1:]:
                            result = _longest_common_substring(result, s)
                        return result

                line = edit.getText()
                suggestions = self.getSuggestions(line)

                if len(suggestions) == 0:
                    return
                if len(suggestions) > 1:
                    self.write('\n'.join(suggestions)+"\n")
                    scrollbar = self.text.verticalScrollBar()
                    scrollbar.setSliderPosition(scrollbar.maximum())
                    edit.setText(_largest_common(suggestions))
                elif len(suggestions) == 1:
                    edit.setText(suggestions[0])

            @self.clear.mhEvent
            def onClicked(event):
                self.clearText()

            @self.line.mhEvent
            def onUpArrow(_dummy):
                self.upArrow()

            @self.line.mhEvent
            def onDownArrow(_dummy):
                self.downArrow()


    def getSuggestions(self, line):
        from rlcompleter import Completer
        def _inRange(i):
            if MAX_COMPLETIONS <= 0:
                # No limit
                return True
            else:
                return i <= MAX_COMPLETIONS
        result = []
        completer = Completer(self.globals)
        i = 0
        suggestion = True
        while suggestion and _inRange(i):
            suggestion = completer.complete(line, i)
            if suggestion:
                if i == MAX_COMPLETIONS and MAX_COMPLETIONS != 0:
                    result.append('...')
                else:
                    if suggestion not in result:
                        result.append(suggestion)
            i += 1
        return result

    def execute(self, text):
        stdout = sys.stdout
        stderr = sys.stderr
        sys.stdout = self
        sys.stderr = self
        try:
            code = compile(text, '<shell>', 'single')
            eval(code, self.globals)
            # exec text in self.globals, {}
        except:
            traceback.print_exc()
        sys.stdout = stdout
        sys.stderr = stderr

    def write(self, text):
        self.text.addText(text)

    def flush(self):
        pass

    def clearText(self):
        self.text.setText('')

    def upArrow(self):
        if not self.history:
            return
        if self.histitem is None or self.histitem == 0:
            self.histitem = len(self.history) - 1
        else:
            self.histitem -= 1
        self.line.setText(self.history[self.histitem])

    def downArrow(self):
        if not self.history:
            return
        if self.histitem is None or self.histitem >= len(self.history) - 1:
            self.histitem = None
            self.line.setText('')
        else:
            self.histitem += 1
            self.line.setText(self.history[self.histitem])

    def onThemeChanged(self, event):
        if self.console is None:
            return
        self.console.set_theme(G.app.theme)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
**Project Name:**      MakeHuman

**Product Home Page:** http://www.makehumancommunity.org/

**Github Code Home Page:**    https://github.com/makehumancommunity/

**Authors:**           Thanasis Papoutsidakis

**Copyright(c):**      MakeHuman Team 2001-2020

**Licensing:**         AGPL3

    This file is part of MakeHuman Community (www.makehumancommunity.org).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.


Abstract
--------

This module contains the Render Task View class to serve as a base class
for task views that implement renderers and rendering related tasks.
"""

from core import G
import gui3d


# TODO does this module still have much use?

class RenderTaskView(gui3d.TaskView):
    def __init__(self, category, name, label=None):
        super(RenderTaskView, self).__init__(category, name, label)

        # Declare settings
        G.app.addSetting('rendering_width', 800)
        G.app.addSetting('rendering_height', 600)

        self.oldShader = None
        self.taskViewShader = None

    # Render task views enable pose mode when shown, so that the selected pose
    # is seen and rendered, and a light-based shader (default None) is
    # selected, so that the actual scene lighting is simulated.

    def onShow(self, event):
        super(RenderTaskView, self).onShow(event)
        import getpath

        human = G.app.selectedHuman
        self.oldShader = human.material.shader
        human.material.shader = getpath.getSysDataPath(self.taskViewShader) if self.taskViewShader else None

    def onHide(self, event):
        human = G.app.selectedHuman
        human.material.shader = self.oldShader

        super(RenderTaskView, self).onHide(event)

    # renderingWidth, renderingHeight: properties for getting/setting
    # the rendering width and height stored in the settings.

    def getRenderingWidth(self):
        return G.app.getSetting('rendering_width')

    def setRenderingWidth(self, value=None):
        G.app.setSetting('rendering_width', 0 if not value else int(value))

    renderingWidth = property(getRenderingWidth, setRenderingWidth)

    def getRenderingHeight(self):
        return G.app.getSetting('rendering_height')

    def setRenderingHeight(self, value=None):
        G.app.setSetting('rendering_height', 0 if not value else int(value))

    renderingHeight = property(getRenderingHeight, setRenderingHeight)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
**Project Name:**      MakeHuman

**Product Home Page:** http://www.makehumancommunity.org/

**Github Code Home Page:**    https://github.com/makehumancommunity/

**Authors:**           Glynn Clements

**Copyright(c):**      MakeHuman Team 2001-2020

**Licensing:**         AGPL3

    This file is part of MakeHuman (www.makehumancommunity.org).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.


Abstract
--------

TODO
"""

#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import pprint
import gui3d
import gui
from core import G
import log

class DataTree(gui.TreeView):
    def __init__(self, root):
        super(DataTree, self).__init__()
        self.item = self.addTopLevel('Application')
        self.root = root

    def getItemPath(self, item):
        path = []
        while item is not None and item.parent is not None:
            path.append(item.text)
            item = item.parent
        path = list(reversed(path))
        return path

    def getValue(self, path):
        root = self.root
        for key in path:
            root = getattr(root, key, None)
            if root is None:
                break
        return root

    def populate(self, item):
        # log.message('populate: %s', item)
        path = self.getItemPath(item)
        # log.message('populate: path=%s', path)
        value = self.getValue(path)
        for name in sorted(value.__dict__.keys()):
            if name[:2] == '__' and name[-2:] == '__':
                continue
            child = getattr(value, name, None)
            if hasattr(child, '__dict__'):
                item.addChild(name, True)
            else:
                item.addChild(name, False)

class DataTaskView(gui3d.TaskView):
    def __init__(self, category):
        super(DataTaskView, self).__init__(category, 'Data')
        self.pp = pprint.PrettyPrinter()
        self.left.child.setSizePolicy(
            gui.SizePolicy.MinimumExpanding,
            gui.SizePolicy.MinimumExpanding)

        self.tree = self.addLeftWidget(DataTree(G.app))
        self.clear = self.addLeftWidget(gui.Button('Clear'))
        self.text = self.addTopWidget(gui.DocumentEdit())

        self.tree.setHeaderHidden(True)
        self.tree.resizeColumnToContents(0)
        self.tree.setSizePolicy(
            gui.SizePolicy.Ignored,
            gui.SizePolicy.Expanding)
        self.left.layout.setStretchFactor(self.tree, 1)

        @self.tree.mhEvent
        def onActivate(item):
            path = self.tree.getItemPath(item)
            # log.message('data: %s', path)
            self.showData(path)

        @self.tree.mhEvent
        def onExpand(item):
            self.tree.populate(item)
            self.tree.resizeColumnToContents(0)

        @self.clear.mhEvent
        def onClicked(event):
            self.clearData()

    def clearData(self):
        self.text.setText('')

    def showData(self, path):
        val = self.tree.getValue(path)
        self.text.setText(self.pp.pformat(val))

def load(app):
    category = app.getCategory('Utilities')
    taskview = category.addTask(DataTaskView(category))

def unload(app):
    pass




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# This file contains a copy of numpy.unique from NumPy 1.6.2
# Earlier versions had a bug, see:
# http://projects.scipy.org/numpy/ticket/2063
# https://github.com/numpy/numpy/commit/dbf235169ed3386b359caaa9217f5280bf1d6749
# If an older version is detected, the numpy module will be monkey-patched to use this version

# Copyright (c) 2005, NumPy Developers
# 
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in the
#     documentation and/or other materials provided with the
#     distribution.
#   * Neither the name of the NumPy Developers nor the names of any
#     contributors may be used to endorse or promote products derived
#     from this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
# IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import numpy as np

def unique(ar, return_index=False, return_inverse=False):
    """
    Find the unique elements of an array.

    Returns the sorted unique elements of an array. There are two optional
    outputs in addition to the unique elements: the indices of the input array
    that give the unique values, and the indices of the unique array that
    reconstruct the input array.

    Parameters
    ----------
    ar : array_like
        Input array. This will be flattened if it is not already 1-D.
    return_index : bool, optional
        If True, also return the indices of `ar` that result in the unique
        array.
    return_inverse : bool, optional
        If True, also return the indices of the unique array that can be used
        to reconstruct `ar`.

    Returns
    -------
    unique : ndarray
        The sorted unique values.
    unique_indices : ndarray, optional
        The indices of the first occurrences of the unique values in the
        (flattened) original array. Only provided if `return_index` is True.
    unique_inverse : ndarray, optional
        The indices to reconstruct the (flattened) original array from the
        unique array. Only provided if `return_inverse` is True.

    See Also
    --------
    numpy.lib.arraysetops : Module with a number of other functions for
                            performing set operations on arrays.

    Examples
    --------
    >>> np.unique([1, 1, 2, 2, 3, 3])
    array([1, 2, 3])
    >>> a = np.array([[1, 1], [2, 3]])
    >>> np.unique(a)
    array([1, 2, 3])

    Return the indices of the original array that give the unique values:

    >>> a = np.array(['a', 'b', 'b', 'c', 'a'])
    >>> u, indices = np.unique(a, return_index=True)
    >>> u
    array(['a', 'b', 'c'],
           dtype='|S1')
    >>> indices
    array([0, 1, 3])
    >>> a[indices]
    array(['a', 'b', 'c'],
           dtype='|S1')

    Reconstruct the input array from the unique values:

    >>> a = np.array([1, 2, 6, 4, 2, 3, 2])
    >>> u, indices = np.unique(a, return_inverse=True)
    >>> u
    array([1, 2, 3, 4, 6])
    >>> indices
    array([0, 1, 4, 3, 1, 2, 1])
    >>> u[indices]
    array([1, 2, 6, 4, 2, 3, 2])

    """
    try:
        ar = ar.flatten()
    except AttributeError:
        if not return_inverse and not return_index:
            items = sorted(set(ar))
            return np.asarray(items)
        else:
            ar = np.asanyarray(ar).flatten()

    if ar.size == 0:
        if return_inverse and return_index:
            return ar, np.empty(0, np.bool), np.empty(0, np.bool)
        elif return_inverse or return_index:
            return ar, np.empty(0, np.bool)
        else:
            return ar

    if return_inverse or return_index:
        if return_index:
            perm = ar.argsort(kind='mergesort')
        else:
            perm = ar.argsort()
        aux = ar[perm]
        flag = np.concatenate(([True], aux[1:] != aux[:-1]))
        if return_inverse:
            iflag = np.cumsum(flag) - 1
            iperm = perm.argsort()
            if return_index:
                return aux[flag], perm[flag], iflag[iperm]
            else:
                return aux[flag], iflag[iperm]
        else:
            return aux[flag], perm[flag]

    else:
        ar.sort()
        flag = np.concatenate(([True], ar[1:] != ar[:-1]))
        return ar[flag]

if np.__version__.split('.') < ['1','6','2']:
    np.unique = unique



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
**Project Name:**      MakeHuman

**Product Home Page:** http://www.makehumancommunity.org/

**Github Code Home Page:**    https://github.com/makehumancommunity/

**Authors:**           Joel Palmius, Aranuvir

**Copyright(c):**      MakeHuman Team 2001-2020

**Licensing:**         AGPL3

    This file is part of MakeHuman (www.makehumancommunity.org).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.


Abstract
--------

TODO
"""
import os
import log
from .savetargets import SaveTargetsTaskView


def load(app):
    category = app.getCategory('Utilities')
    taskview = category.addTask(SaveTargetsTaskView(category))


def unload(app):

    category = app.getCategory('Utilities')
    taskview = category.getTaskByName('Save Targets')


    if os.path.isfile(taskview.metaFile):
        try:
            os.remove(taskview.metaFile)
        except OSError as e:
            log.error('Cannot delete meta target file: %s',  format(taskview.metaFile))

    if os.path.isdir(taskview.metaFilePath):
        try:
            os.rmdir(taskview.metaFilePath)
        except OSError as e:
            if e.errno == 39:
                log.warning('Cannot delete save targets cache: %s. Folder still in use...', format(taskview.metaFilePath))
            else:
                log.error('Cannot delete save targets cache: %s', format(taskview.metaFilePath))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
**Project Name:**      MakeHuman

**Product Home Page:** http://www.makehumancommunity.org/

**Github Code Home Page:**    https://github.com/makehumancommunity/

**Authors:**           Jonas Hauquier

**Copyright(c):**      MakeHuman Team 2001-2020

**Licensing:**         AGPL3

    This file is part of MakeHuman (www.makehumancommunity.org).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.


Abstract
--------

Exporter plugin for the Ogre3d mesh format.
"""

from . import mh2ogre
from export import Exporter, ExportConfig
import imp

class OgreConfig(ExportConfig):

    def __init__(self):
        ExportConfig.__init__(self)
        self.useRelPaths = True
        self.exportShaders = False  # TODO add support for this

    @property
    def subdivide(self):
        return self.human.isSubdivided()

class ExporterOgre(Exporter):
    def __init__(self):
        Exporter.__init__(self)
        self.name = "Ogre3D"
        self.filter = "Ogre3D Mesh XML (*.mesh.xml)"
        self.fileExtension = "mesh.xml"
        self.orderPriority = 60.0

    def export(self, human, filename):
        imp.reload(mh2ogre) # TODO ?
        cfg = self.getConfig()
        cfg.setHuman(human)
        mh2ogre.exportOgreMesh(filename("mesh.xml"), cfg)

    def build(self, options, taskview):
        import gui
        self.taskview     = taskview
        self.feetOnGround = options.addWidget(gui.CheckBox("Feet on ground", True))

    def getConfig(self):
        cfg = OgreConfig()
        cfg.feetOnGround      = self.feetOnGround.selected
        cfg.scale,cfg.unit    = self.taskview.getScale()

        return cfg

def load(app):
    app.addExporter(ExporterOgre())

def unload(app):
    pass




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
**Project Name:**      MakeHuman

**Product Home Page:** http://www.makehumancommunity.org/

**Github Code Home Page:**    https://github.com/makehumancommunity/

**Authors:**           Glynn Clements, Jonas Hauquier

**Copyright(c):**      MakeHuman Team 2001-2020

**Licensing:**         AGPL3

    This file is part of MakeHuman Community (www.makehumancommunity.org).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.


Abstract
--------

TODO
"""

import sys
import os
import re
from core import G
import log

from PyQt5 import QtCore, QtGui, QtSvg, QtWidgets

import events3d
import language
#import log
from getpath import getSysDataPath, getPath, isSubPath, pathToUnicode

def dummySvgCall():
    """Code which is here just so pyinstaller can discover we need SVG support"""
    dummy = QtSvg.QGraphicsSvgItem("some_svg.svg")
    
def getLanguageString(text, appendData=None, appendFormat=None):
    """Function to get the translation of a text according to the selected
    language.

    The function will look up the given text in the current language's
    dictionary and it will return the translated string.
    """
    if not text:
        return text
    return language.language.getLanguageString(text,appendData,appendFormat)


class Widget(events3d.EventHandler):
    def __init__(self):
        events3d.EventHandler.__init__(self)

    def callEvent(self, eventType, event):
        super(Widget, self).callEvent(eventType, event)

    def focusInEvent(self, event):
        self.callEvent('onFocus', self)
        super(type(self), self).focusInEvent(event)

    def focusOutEvent(self, event):
        self.callEvent('onBlur', self)
        super(type(self), self).focusOutEvent(event)

    def showEvent(self, event):
        self.callEvent('onShow', self)
        super(type(self), self).showEvent(event)

    def hideEvent(self, event):
        self.callEvent('onHide', self)
        super(type(self), self).hideEvent(event)

    def onFocus(self, event):
        pass

    def onBlur(self, event):
        pass

    def onShow(self, event):
        pass

    def onHide(self, event):
        pass

class Tab(Widget):
    def __init__(self, parent, name, label):
        super(Tab, self).__init__()
        self.parent = parent
        self.name = name
        self.label = label

    def onClicked(self, event):
        pass

class TabsBase(Widget):
    def __init__(self):
        super(TabsBase, self).__init__()
        self.tabBar().setExpanding(False)
        self.currentChanged[int].connect(self.tabChanged)
        self._tabs_by_idx = {}
        self._tabs_by_name = {}

    def _addTab(self, name, label, idx=None):
        label = getLanguageString(label)
        tab = Tab(self, name, label)
        tab.idx = self._makeTab(tab, idx)
        if idx != None:
            # Update index list when inserting tabs at arbitrary positions
            newIdxList = {}
            for tIdx, t in list(self._tabs_by_idx.items()):
                if int(tIdx) >= idx:
                    t.idx += 1
                newIdxList[t.idx] = t
            self._tabs_by_idx = newIdxList
        self._tabs_by_idx[tab.idx] = tab
        self._tabs_by_name[tab.name] = tab
        return tab

    def tabChanged(self, idx):
        tab = self._tabs_by_idx.get(idx)
        if tab:
            self.callEvent('onTabSelected', tab)
            tab.callEvent('onClicked', tab)

    def findTab(self, name):
        return self._tabs_by_name.get(name)

    def changeTab(self, name):
        tab = self.findTab(name)
        if tab is None:
            return
        self.setCurrentIndex(tab.idx)

    def onTabSelected(self, event):
        pass

class Tabs(QtWidgets.QTabWidget, TabsBase):
    def __init__(self, parent = None):
        QtWidgets.QTabWidget.__init__(self, parent)
        #TabsBase.__init__(self)

    def _makeTab(self, tab, idx=None):
        tab.child = TabBar(self)
        if idx != None:
            i = super(Tabs, self).insertTab(idx, tab.child, tab.label)
            if i == 0:
                self.setCurrentIndex(0)
            return i
        return super(Tabs, self).addTab(tab.child, tab.label)

    def addTab(self, name, label, idx = None):
        return super(Tabs, self)._addTab(name, label, idx)

    def tabChanged(self, idx):
        super(Tabs, self).tabChanged(idx)
        tab = self._tabs_by_idx.get(idx)
        if tab:
            tab.child.tabChanged(tab.child.currentIndex())

class TabBar(QtWidgets.QTabBar, TabsBase):
    def __init__(self, parent = None):
        QtWidgets.QTabBar.__init__(self, parent)
        #TabsBase.__init__(self)
        self.setDrawBase(False)

    def tabBar(self):
        return self

    def _makeTab(self, tab, idx = None):
        if idx != None:
            i = super(TabBar, self).insertTab(idx, tab.label)
            if i == 0:
                self.setCurrentIndex(0)
            return i
        return super(TabBar, self).addTab(tab.label)

    def addTab(self, name, label, idx = None):
        return super(TabBar, self)._addTab(name, label, idx)

class GroupBox(QtWidgets.QGroupBox, Widget):
    def __init__(self, label = ''):
        label = getLanguageString(label) if label else ''
        QtWidgets.QGroupBox.__init__(self, label)
        #Widget.__init__(self)
        self.layout = QtWidgets.QGridLayout(self)

    def __str__(self):
        return "%s - %s" % (type(self), str(self.title()))

    def addWidget(self, widget, row = None, column = 0, rowSpan = 1, columnSpan = 1, alignment = QtCore.Qt.Alignment(0)):
        # widget.setParent(self)
        if row is None:
            row = self.layout.count()
        self.layout.addWidget(widget, row, column, rowSpan, columnSpan, alignment)
        widget.show()
        return widget

    def removeWidget(self, widget):
        self.layout.removeWidget(widget)
        widget.setParent(None)

    @property
    def children(self):
        return list(self.layout.itemAt(i).widget() for i in range(self.count()))

    def count(self):
        return self.layout.count()

    def itemAt(self, idx):
        return self.layout.itemAt(idx)

# PyQt doesn't implement QProxyStyle so we have to do all this ...

class SliderStyle(QtWidgets.QCommonStyle):
    def __init__(self, parent):
        self.__parent = parent
        super(SliderStyle, self).__init__()

    def drawComplexControl(self, control, option, painter, widget = None):
        return self.__parent.drawComplexControl(control, option, painter, widget)

    def drawControl(self, element, option, painter, widget = None):
        return self.__parent.drawControl(element, option, painter, widget)

    def drawItemPixmap(self, painter, rectangle, alignment, pixmap):
        return self.__parent.drawItemPixmap(painter, rectangle, alignment, pixmap)

    def drawItemText(self, painter, rectangle, alignment, palette, enabled, text, textRole = QtGui.QPalette.NoRole):
        return self.__parent.drawItemText(painter, rectangle, alignment, palette, enabled, text, textRole)

    def drawPrimitive(self, element, option, painter, widget = None):
        return self.__parent.drawPrimitive(element, option, painter, widget)

    def generatedIconPixmap(self, iconMode, pixmap, option):
        return self.__parent.generatedIconPixmap(iconMode, pixmap, option)

    def hitTestComplexControl(self, control, option, position, widget = None):
        return self.__parent.hitTestComplexControl(control, option, position, widget)

    def itemPixmapRect(self, rectangle, alignment, pixmap):
        return self.__parent.itemPixmapRect(rectangle, alignment, pixmap)

    def itemTextRect(self, metrics, rectangle, alignment, enabled, text):
        return self.__parent.itemTextRect(metrics, rectangle, alignment, enabled, text)

    def pixelMetric(self, metric, option = None, widget = None):
        return self.__parent.pixelMetric(metric, option, widget)

    def polish(self, *args, **kwargs):
        return self.__parent.polish(*args, **kwargs)

    def styleHint(self, hint, option=None, widget=None, returnData=None):
        if hint == QtWidgets.QStyle.SH_Slider_AbsoluteSetButtons:
            return QtCore.Qt.LeftButton | QtCore.Qt.MidButton | QtCore.Qt.RightButton
        return self.__parent.styleHint(hint, option, widget, returnData)

    def subControlRect(self, control, option, subControl, widget = None):
        return self.__parent.subControlRect(control, option, subControl, widget)

    def subElementRect(self, element, option, widget = None):
        return self.__parent.subElementRect(element, option, widget)

    def unpolish(self, *args, **kwargs):
        return self.__parent.unpolish(*args, **kwargs)

    def sizeFromContents(self, ct, opt, contentsSize, widget = None):
        return self.__parent.sizeFromContents(ct, opt, contentsSize, widget)

class NarrowLineEdit(QtWidgets.QLineEdit):
    def __init__(self, width=4, *args, **kwargs):
        super(NarrowLineEdit, self).__init__(*args, **kwargs)
        self.__cols = width

    def sizeHint(self):
        self.ensurePolished()
        fm = QtGui.QFontMetrics(self.font())
        leftMargin, topMargin, rightMargin, bottomMargin = self.getContentsMargins()
        textMargins = self.textMargins()
        h = max(fm.height(), 14) + 2 + textMargins.top() + textMargins.bottom() + topMargin + bottomMargin
        w = fm.width('0') * self.__cols + 4 + textMargins.left() + textMargins.right() + leftMargin + rightMargin

        opt = QtWidgets.QStyleOptionFrame()
        self.initStyleOption(opt)
        return self.style().sizeFromContents(
            QtWidgets.QStyle.CT_LineEdit, opt,
            QtCore.QSize(w, h).expandedTo(QtWidgets.QApplication.globalStrut()),
            self)

class _QSlider(QtWidgets.QSlider):
    """
    Mock object around QSlider that allows catching mouse press events and
    relaying them to the parent widget.
    """

    def __init__(self, parent, orientation):
        super(_QSlider, self).__init__(orientation)
        self.parent = parent

    def mousePressEvent(self, event):
        if self.parent:
            if not self.parent.sliderMousePressEvent(event):
                return
        super(_QSlider, self).mousePressEvent(event)

class Slider(QtWidgets.QWidget, Widget):
    _imageCache = {}
    _show_images = False
    _instances = set()
    _style = None

    @classmethod
    def _getImage(cls, path):
        if path not in cls._imageCache:
            cls._imageCache[path] = getPixmap(path)
        return cls._imageCache[path]

    def __init__(self, value=0.0, min=0.0, max=1.0, label=None, vertical=False, valueConverter=None, image=None, scale=1000):
        super(Slider, self).__init__()
        #Widget.__init__(self)
        self.text = getLanguageString(label) or ''
        self.valueConverter = valueConverter

        orient = (QtCore.Qt.Vertical if vertical else QtCore.Qt.Horizontal)
        self.slider = _QSlider(self, orient)
        if Slider._style is None:
            Slider._style = SliderStyle(self.slider.style())
        self.slider.setStyle(Slider._style)

        self.min = min
        self.max = max
        self.scale = scale
        self.slider.setMinimum(0)
        self.slider.setMaximum(self.scale)
        self.slider.setValue(self._f2i(value))
        self.slider.setTracking(False)
        self.slider.sliderMoved[int].connect(self._changing)
        self.slider.valueChanged[int].connect(self._changed)
        self.slider.sliderReleased.connect(self._released)
        self.slider.sliderPressed.connect(self._pressed)
        self.slider.installEventFilter(self)

        self.label = QtWidgets.QLabel(self.text)
        # Decrease vertical gap between label and slider
        #self.label.setContentsMargins(0, 0, 0, -1)
        self.layout = QtWidgets.QGridLayout(self)
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.layout.setColumnMinimumWidth(1, 1)
        self.layout.setColumnStretch(0, 1)
        self.layout.setColumnStretch(1, 0)
        self.layout.setColumnStretch(2, 0)
        self.layout.addWidget(self.label, 1, 0, 1, 1)
        self.layout.addWidget(self.slider, 2, 0, 1, -1)
        if not self.text:
            self.label.hide()

        if image is not None:
            self.image = QtWidgets.QLabel()
            self.image.setPixmap(self._getImage(image))
            self.layout.addWidget(self.image, 0, 0, 1, -1)
        else:
            self.image = None

        self._sync(value)
        self._update_image()

        type(self)._instances.add(self)

    def setValueConverter(self, valueConverter):
        self._valueConverter = valueConverter
        if self.valueConverter:
            self.edit = NarrowLineEdit(5)
            self.edit.installEventFilter(self)
            self.edit.returnPressed.connect(self._enter)
            self.layout.addWidget(self.edit, 1, 1, 1, 1)
            if hasattr(self.valueConverter, 'units'):
                self.units = QtWidgets.QLabel(self.valueConverter.units)
                self.layout.addWidget(self.units, 1, 2, 1, 1)
            else:
                self.units = None
        else:
            self.edit = None
            self.units = None

    def getValueConverter(self):
        return self._valueConverter

    valueConverter = property(getValueConverter, setValueConverter)

    def sliderMousePressEvent(self, event):
        """
        Can be used to catch mouse press events from the attached QSlider widget.
        Return True to execute the default behaviour of the slider widget,
        return False to interrupt further handling of the event.
        Override this method when needed.
        """
        return True

    def __del__(self):
        type(self)._instances.remove(self)

    def _enter(self):
        text = str(self.edit.text())
        if not text:
            return
        oldValue = self.getValue()
        newValue = self.fromDisplay(float(text))
        self.setValue(newValue)
        if abs(oldValue - newValue) > 1e-3:
            self.callEvent('onChange', newValue)

    def toDisplay(self, value):
        if self.valueConverter is None:
            return value
        else:
            return self.valueConverter.dataToDisplay(value)

    def fromDisplay(self, value):
        if self.valueConverter is None:
            return value
        else:
            return self.valueConverter.displayToData(value)

    def eventFilter(self, object, event):
        if object != self.slider and object != self.edit:
            return
        result = False
        if object == self.edit:
            result = self.edit.eventFilter(object, event)
        if event.type() == QtCore.QEvent.FocusIn:
            self.callEvent('onFocus', self)
        elif event.type() == QtCore.QEvent.FocusOut:
            self.callEvent('onBlur', self)
        return result

    def _update_image(self):
        if self.image is None:
            return
        if type(self)._show_images:
            self.image.show()
        else:
            self.image.hide()

    @classmethod
    def imagesShown(cls):
        return cls._show_images

    @classmethod
    def showImages(cls, state):
        cls._show_images = state
        for w in cls._instances:
            w._update_image()

    def _changing(self, value):
        value = self._i2f(value)
        self._sync(value)
        self.callEvent('onChanging', value)

    def _changed(self, value):
        value = self._i2f(value)
        self._sync(value)
        self.callEvent('onChange', value)

    def _released(self):
        self.callEvent('onRelease', self)

    def _pressed(self):
        self.callEvent('onPress', self)

    def _sync(self, value):
        if '%' in self.text:
            self.label.setText(self.text % self.toDisplay(value))
        if self.edit is not None:
            self.edit.setText('%.2f' % self.toDisplay(value))
        if hasattr(self.valueConverter, 'units') and \
           self.valueConverter.units != str(self.units.text()):
            self.units.setText(self.valueConverter.units)

    def _f2i(self, x):
        return int(round(self.scale * (x - self.min) / (self.max - self.min)))

    def _i2f(self, x):
        return self.min + (x / float(self.scale)) * (self.max - self.min)

    def setValue(self, value):
        vmax = max(self.max, self.min)  # Virtual min and max values.
        vmin = min(self.max, self.min)  # Useful in case the slider is reversed.
        value = min(vmax, max(vmin, value))
        self._sync(value)
        if self._f2i(value) == self.slider.value():
            return
        self.slider.blockSignals(True)
        self.slider.setValue(self._f2i(value))
        self.slider.blockSignals(False)

    def getValue(self):
        return self._i2f(self.slider.value())

    def setMin(self, min):
        value = self.getValue()
        self.min = min
        self.setValue(value)

    def setMax(self, max):
        value = self.getValue()
        self.max = max
        self.setValue(value)

    def onChanging(self, event):
        pass

    def onChange(self, event):
        pass

class ButtonBase(Widget):
    def __init__(self):
        Widget.__init__(self)
        self.clicked[bool].connect(self._clicked)

    def getLabel(self):
        return str(self.text())

    def setLabel(self, label):
        label = getLanguageString(label)
        self.setText(label)

    def _clicked(self, state):
        self.callEvent('onClicked', None)

    @property
    def selected(self):
        return self.isChecked()

    def setSelected(self, value):
        self.setChecked(value)

    def onClicked(self, event):
        pass

class Button(QtWidgets.QPushButton, ButtonBase):
    def __init__(self, label=None, selected=False):
        label = getLanguageString(label)
        super(Button, self).__init__(label)
        #ButtonBase.__init__(self)

class CheckBox(QtWidgets.QCheckBox, ButtonBase):
    def __init__(self, label=None, selected=False):
        label = getLanguageString(label)
        super(CheckBox, self).__init__(label)
        #ButtonBase.__init__(self)
        self.setChecked(selected)

class RadioButton(QtWidgets.QRadioButton, ButtonBase):
    groups = {}

    def __init__(self, group, label=None, selected=False):
        label = getLanguageString(label)
        super(RadioButton, self).__init__(label)
        #ButtonBase.__init__(self)
        self.group = group
        self.group.append(self)
        self.setChecked(selected)
        self._addToGroup(group)

    #def __del__(self):
    #    self._removeFromGroup(self.group)

    def _addToGroup(self, group):
        if id(group) in type(self).groups:
            rbgroup = type(self).groups[id(group)]
        else:
            rbgroup = QtWidgets.QButtonGroup()
            rbgroup.setExclusive(True)
            type(self).groups[id(group)] = rbgroup
        rbgroup.addButton(self)

    def _removeFromGroup(self, group):
        if id(group) not in type(self).groups:
            return
        rbgroup = type(self).groups[id(group)]
        rbgroup.removeButton(self)
        if len(rbgroup.buttons()) == 0:
            del type(self).groups[id(group)]

    @property
    def selected(self):
        return self.isChecked()

    def getSelection(self):
        for radio in self.group:
            if radio.selected:
                return radio

class ListItem(QtWidgets.QListWidgetItem):
    def __init__(self, label, tooltip = True):
        super(ListItem, self).__init__(label)
        self.__hasCheckbox = False
        self.tooltip = tooltip

    def updateTooltip(self):
        """
        Attach a mouse-over tooltip for this item if the text is too long to fit
        the widget.
        """
        if not self.tooltip:
            return

        if not self.listWidget():
            return

        metrics = QtGui.QFontMetrics(self.font())

        labelWidth = self.listWidget().width()
        if self.icon():
            labelWidth -= self.listWidget().iconSize().width() + 10
            # pad size with 10px to account for margin between icon and text (this is an approximation)

        if metrics.width(self.text)//2 > labelWidth:
            self.setToolTip(self.text)
        else:
            self.setToolTip("")

    @property
    def hasCheckbox(self):
        return self.__hasCheckbox

    def setUserData(self, data):
        self.setData(QtCore.Qt.UserRole, data)

    def getUserData(self):
        return self.data(QtCore.Qt.UserRole)

    def setText(self, text):
        super(ListItem, self).setText(text)
        self.updateTooltip()

    @property
    def text(self):
        return str(super(ListItem, self).text())

    def enableCheckbox(self):
        self.__hasCheckbox = True
        self.setFlags(self.flags() | QtCore.Qt.ItemIsUserCheckable)
        self.setCheckState(QtCore.Qt.Unchecked)
        self.checkedState = False

    def setChecked(self, checked):
        if not self.hasCheckbox:
            self.setSelected(checked)
            return

        if checked:
            self.setCheckState(QtCore.Qt.Checked)
        else:
            self.setCheckState(QtCore.Qt.Unchecked)
        self.checkedState = self.checkState()

    def isChecked(self):
        return self.hasCheckbox and self.checkState() != QtCore.Qt.Unchecked

    def _clicked(self):
        owner = self.listWidget()
        if self.hasCheckbox:
            if self.checkState() != self.checkedState:
                self.checkedState = self.checkState()
                if self.checkState():
                    owner.callEvent('onItemChecked', self)
                else:
                    owner.callEvent('onItemUnchecked', self)
                return True
        return False

class ListView(QtWidgets.QListWidget, Widget):
    def __init__(self):
        super(ListView, self).__init__()
        #Widget.__init__(self)
        self._vertical_scrolling = True
        self.itemActivated[QtWidgets.QListWidgetItem].connect(self._activate)
        self.itemClicked[QtWidgets.QListWidgetItem].connect(self._clicked)

    def _activate(self, item):
        self.callEvent('onActivate', item)

    def _clicked(self, item):
        if item._clicked():
            return
        if self.allowsMultipleSelection():
            if item.isSelected():
                self.callEvent('onItemChecked', item)
            else:
                self.callEvent('onItemUnchecked', item)
        else:
            self.callEvent('onClicked', item)

    def onActivate(self, event):
        pass

    def onClicked(self, event):
        pass

    def setData(self, items):
        self.clear()
        for item in items:
            self.addItem(item)

    def setVerticalScrollingEnabled(self, enabled):
        self._vertical_scrolling = enabled
        if enabled:
            self.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarAsNeeded)
        else:
            self.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)
        self.updateGeometry()

    def rowCount(self):
        return len( [item for item in self.getItems() if not item.isHidden()] )

    def sizeHint(self):
        if self._vertical_scrolling:
            return super(ListView, self).sizeHint()
        else:
            rows = self.rowCount()
            if rows > 0:
                rowHeight = self.sizeHintForRow(0)
                rowHeight = max(rowHeight, self.iconSize().height())
            else:
                rowHeight = 0
            height = (rowHeight * rows)

            size = super(ListView,self).sizeHint()
            size.setHeight(height)
            return size

    _brushes = {}
    @classmethod
    def getBrush(cls, color):
        if color not in cls._brushes:
            cls._brushes[color] = QtGui.QBrush(QtGui.QColor(color))
        return cls._brushes[color]

    def addItem(self, text, color = None, data = None, checkbox = False, pos = None):
        item = ListItem(text)
        item.setText(text)
        if color is not None:
            item.setForeground(self.getBrush(color))
        if data is not None:
            item.setUserData(data)
        if checkbox:
            item.enableCheckbox()
        return self.addItemObject(item, pos)

    def addItemObject(self, item, pos = None):
        if pos is not None:
            super(ListView, self).insertItem(pos, item)
        else:
            super(ListView, self).addItem(item)
        if not self._vertical_scrolling:
            self.updateGeometry()
        return item

    def getSelectedItem(self):
        items = self.selectedItems()
        if len(items) > 0:
            return items[0].text
        return None

    def getSelectedItems(self):
        return [item.text for item in self.selectedItems()]

    def getItemData(self, row):
        return self.item(row).getUserData()

    def setItemColor(self, row, color):
        self.item(row).setForeground(self.getBrush(color))

    def showItem(self, row, state):
        self.item(row).setHidden(not state)

    def allowMultipleSelection(self, allow):
        self.setSelectionMode(QtWidgets.QAbstractItemView.MultiSelection
                              if allow else
                              QtWidgets.QAbstractItemView.SingleSelection)

    def allowsMultipleSelection(self):
        return self.selectionMode() == QtWidgets.QAbstractItemView.MultiSelection

    def getItems(self):
        return [ self.item(row) for row in range(self.count()) ]

    def clearSelection(self):
        super(ListView, self).clearSelection()

        for item in self.getItems():
            if item.isChecked():
                item.setChecked(False)

        self.callEvent('onClearSelection', None)

class TextView(QtWidgets.QLabel, Widget):
    def __init__(self, label = ''):
        label = getLanguageString(label)
        super(TextView, self).__init__(label)
        #Widget.__init__(self)

    def setText(self, text):
        text = getLanguageString(text)
        super(TextView,self).setText(text)

    def setTextFormat(self, text, *values):
        text = getLanguageString(text)
        super(TextView,self).setText(text % values)

    # Workaround for incorrect resizing with enabled word wrapping
    # https://bugreports.qt.io/browse/QTBUG-37673
    def resizeEvent(self, event):
        super(TextView, self).resizeEvent(event)

        if self.wordWrap():
            # heightForWidth rely on minimumSize to evaulate, so reset it before
            self.setMinimumHeight(0)
            # define minimum height
            self.setMinimumHeight(self.heightForWidth(self.width()))

class SliderBox(GroupBox):
    pass

def intValidator(text):
    return not text or text.isdigit() or (text[0] == '-' and (len(text) == 1 or text[1:].isdigit()))

def floatValidator(text):
    return not text or (text.replace('.', '').isdigit() and text.count('.') <= 1) or (text[0] == '-' and (len(text) == 1 or text[1:].replace('.', '').isdigit()) and text.count('.') <= 1) # Negative sign and optionally digits with optionally 1 decimal point

def filenameValidator(text):
    return not text or len(set(text) & set('\\/:*?"<>|')) == 0

class TextEdit(QtWidgets.QLineEdit, Widget):
    def __init__(self, text='', validator = None):
        super(TextEdit, self).__init__(text)
        #Widget.__init__(self)
        self.setValidator(validator)
        self.textEdited['QString'].connect(self._textChanged)
        self.returnPressed.connect(self._enter)
        self.installEventFilter(self)

    @property
    def text(self):
        return self.getText()

    def _textChanged(self, string):
        self.callEvent('onChange', string)

    def _enter(self):
        self.callEvent('onActivate', self.getText())

    def setText(self, text):
        super(TextEdit, self).setText(text)
        self.setCursorPosition(len(text))

    def getText(self):
        return str(super(TextEdit, self).text())

    def validateText(self, text):
        if self.__validator:
            return self.__validator(text)
        else:
            return True

    def setValidator(self, validator):
        self.__validator = validator
        if validator == intValidator:
            qvalidator = QtGui.QIntValidator()
        elif validator == floatValidator:
            qvalidator = QtGui.QDoubleValidator()
        elif validator == filenameValidator:
            qvalidator = QtGui.QRegExpValidator(QtCore.QRegExp(r'[^\/:*?"<>|]*'))
        else:
            qvalidator = None
        super(TextEdit, self).setValidator(qvalidator)

    def onChange(self, event):
        pass

    def eventFilter(self, object, event):
        if (object is self
            and event.type() == QtCore.QEvent.ShortcutOverride
            and event.key() in (QtCore.Qt.Key_Up, QtCore.Qt.Key_Down)):
            event.accept()
            return True
        return False

    def keyPressEvent(self, event):
        key = event.key()
        if key == QtCore.Qt.Key_Up:
            self._key_up()
            event.accept()
        elif key == QtCore.Qt.Key_Down:
            self._key_down()
            event.accept()
        elif key == QtCore.Qt.Key_Tab and self.tabPressed():
            event.accept()
        else:
            mod = int(event.modifiers())
            if mod > 0:
                self.callEvent('onModifier', (mod, key))
            super(TextEdit, self).keyPressEvent(event)

    def event(self, event):
        if event.type() == QtCore.QEvent.KeyPress and \
           event.key() == QtCore.Qt.Key_Tab and \
           self.tabPressed():
                return True

        return super(TextEdit, self).event(event)

    def tabPressed(self):
        """
        Override and return True to override custom behaviour when TAB key
        is pressed
        """
        return False  # Continue default event handling

    def _key_up(self):
        self.callEvent('onUpArrow', None)

    def _key_down(self):
        self.callEvent('onDownArrow', None)

class DocumentEdit(QtWidgets.QTextEdit, Widget):
    NoWrap          = QtWidgets.QTextEdit.NoWrap
    WidgetWidth     = QtWidgets.QTextEdit.WidgetWidth
    FixedPixelWidth = QtWidgets.QTextEdit.FixedPixelWidth
    FixedColumnWidth= QtWidgets.QTextEdit.FixedColumnWidth

    def __init__(self, text=''):
        super(DocumentEdit, self).__init__(text)
        #Widget.__init__(self)
        self.setAcceptRichText(False)

    @property
    def text(self):
        return self.getText()

    def setText(self, text):
        self.setPlainText(text)

    def addText(self, text):
        self.moveCursor(QtGui.QTextCursor.End, QtGui.QTextCursor.MoveAnchor)
        self.insertPlainText(text)

    def getText(self):
        return str(super(DocumentEdit, self).toPlainText())

class ProgressBar(QtWidgets.QProgressBar, Widget):
    def __init__(self, visible=True):
        super(ProgressBar, self).__init__()
        #Widget.__init__(self)
        self.setVisible(visible)

    def setProgress(self, progress):
        min_ = self.minimum()
        max_ = self.maximum()
        self.setValue(min_ + progress * (max_ - min_))

class ShortcutEdit(QtWidgets.QLabel, Widget):
    def __init__(self, shortcut):
        if shortcut is not None:
            modifiers, key = shortcut
            text = self.shortcutToLabel(modifiers, key)
        else:
            text = ''
        super(ShortcutEdit, self).__init__(text)
        self.setAutoFillBackground(True)
        self.setFocusPolicy(QtCore.Qt.StrongFocus)
        self.setFrameStyle(QtWidgets.QFrame.Panel | QtWidgets.QFrame.Raised)
        self.installEventFilter(self)

    def onFocus(self, arg):
        self.setBackgroundRole(QtGui.QPalette.Highlight)
        self.setForegroundRole(QtGui.QPalette.HighlightedText)

    def onBlur(self, arg):
        self.setBackgroundRole(QtGui.QPalette.Window)
        self.setForegroundRole(QtGui.QPalette.WindowText)

    def setShortcut(self, shortcut):
        modifiers, key = shortcut
        self.setText(self.shortcutToLabel(modifiers, key))

    def eventFilter(self, object, event):
        if object is self and event.type() == QtCore.QEvent.ShortcutOverride:
            event.accept()
            return True
        return False

    def keyPressEvent(self, event):
        key = event.key()
        mod = int(event.modifiers()) & ~QtCore.Qt.ShiftModifier
        if key in (QtCore.Qt.Key_Shift, QtCore.Qt.Key_Control, QtCore.Qt.Key_Alt, QtCore.Qt.Key_Meta):
            return
        self.setText(self.shortcutToLabel(mod, key))
        self.callEvent('onChanged', (mod, key))
        event.accept()

    def shortcutToLabel(self, mod, key):
        mod &= ~0x20000000 # Qt Bug #4022
        seq = QtGui.QKeySequence(key + mod)
        s = str(seq.toString(QtGui.QKeySequence.NativeText))
        return s

    def onChanged(self, shortcut):
        pass

class MouseActionEdit(QtWidgets.QLabel, Widget):
    def __init__(self, shortcut):
        modifiers, button = shortcut
        text = self.shortcutToLabel(modifiers, button)
        super(MouseActionEdit, self).__init__(text)
        self.setFrameStyle(QtWidgets.QFrame.Panel | QtWidgets.QFrame.Raised)

    def setShortcut(self, shortcut):
        modifiers, button = shortcut
        self.setText(self.shortcutToLabel(modifiers, button))

    def mousePressEvent(self, event):
        button = event.button()
        modifiers = int(event.modifiers())
        self.setText(self.shortcutToLabel(modifiers, button))
        self.callEvent('onChanged', (modifiers, button))

    def shortcutToLabel(self, modifiers, button):
        labels = []

        if modifiers & QtCore.Qt.ControlModifier:
            labels.append('Ctrl')
        if modifiers & QtCore.Qt.AltModifier:
            labels.append('Alt')
        if modifiers & QtCore.Qt.MetaModifier:
            labels.append('Meta')
        if modifiers & QtCore.Qt.ShiftModifier:
            labels.append('Shift')

        if button & QtCore.Qt.LeftButton:
            labels.append('Left')
        elif button & QtCore.Qt.MidButton:
            labels.append('Middle')
        elif button & QtCore.Qt.RightButton:
            labels.append('Right')
        else:
            labels.append('[Unknown]')

        return '+'.join(labels)

    def onChanged(self, shortcut):
        pass

class StackedBox(QtWidgets.QStackedWidget, Widget):
    def __init__(self):
        super(StackedBox, self).__init__()
        #Widget.__init__(self)
        self.layout().setAlignment(QtCore.Qt.AlignTop)
        self.autoResize = False
        self.currentChanged[int].connect(self.widgetChanged)

    def addWidget(self, widget):
        w = QtWidgets.QWidget()
        layout = QtWidgets.QVBoxLayout(w)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(widget)
        layout.addStretch()
        super(StackedBox, self).addWidget(w)
        self._updateSize()
        return widget

    def removeWidget(self, widget):
        w = widget.parentWidget()
        super(StackedBox, self).removeWidget(w)
        w.layout().removeWidget(widget)
        w.destroy()
        self._updateSize()

    def showWidget(self, widget):
        self.setCurrentWidget(widget.parentWidget())

    def setAutoResize(self, enabled):
        """
        Set to true to enable auto resizing the vertical dimensions of this
        widget to the height of the currently shown widget.
        """
        self.autoResize = enabled
        self._updateSize()

    def _updateSize(self):
        if self.autoResize:
            for i in range(self.count()):
                w = self.widget(i)
                if w: w.setSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Ignored)
            w = self.widget(self.currentIndex())
            if w: w.setSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        else:
            for i in range(self.count()):
                w = self.widget(i)
                if w: w.setSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        self.layout().activate()

    def widgetChanged(self, widgetIdx):
        self._updateSize()

class Dialog(QtWidgets.QDialog):
    def __init__(self, parent = None):
        super(Dialog, self).__init__(parent)
        self.setModal(True)

        self.helpIds = set()

        icon = self.style().standardIcon(QtWidgets.QStyle.SP_MessageBoxWarning)

        self.layout = QtWidgets.QGridLayout(self)
        self.layout.setColumnStretch(0, 0)
        self.layout.setColumnStretch(1, 1)
        self.layout.setColumnStretch(2, 0)
        self.layout.setColumnStretch(3, 0)

        self.icon = QtWidgets.QLabel()
        self.icon.setPixmap(icon.pixmap(64))
        self.layout.addWidget(self.icon, 0, 0, 2, 1)

        self.text = TextView()
        self.layout.addWidget(self.text, 0, 1, 1, -1)

        self.check = QtWidgets.QCheckBox(getLanguageString("Don't show this again"))
        self.layout.addWidget(self.check, 1, 1, 1, -1)

        self.button1 = QtWidgets.QPushButton()
        self.layout.addWidget(self.button1, 2, 2)

        self.button2 = QtWidgets.QPushButton()
        self.layout.addWidget(self.button2, 2, 3)

        self.button1.clicked[bool].connect(self.accept)
        self.button2.clicked[bool].connect(self.reject)

    def prompt(self, title, text, button1Label, button2Label=None, button1Action=None, button2Action=None, helpId=None, fmtArgs = None):
        if helpId in self.helpIds:
            return

        if fmtArgs is None:
            fmtArgs = []
        elif isinstance(fmtArgs, str):
            fmtArgs = [fmtArgs]

        button1Label = getLanguageString(button1Label)
        button2Label = getLanguageString(button2Label)
        title = getLanguageString(title)

        self.setWindowTitle(title)
        self.text.setTextFormat(text, *fmtArgs)
        self.button1.setText(button1Label)

        if button2Label is not None:
            self.button2.setText(button2Label)
            self.button2.show()
        else:
            self.button2.hide()

        if helpId:
            self.check.show()
            self.check.setChecked(False)
        else:
            self.check.hide()

        which = self.exec_()

        if helpId and self.check.isChecked():
            self.helpIds.add(helpId)

        if which == QtWidgets.QDialog.Accepted:
            if button1Action:
                button1Action()
            return True
        elif which == QtWidgets.QDialog.Rejected:
            if button2Action:
                button2Action()
            return False


class AboutBox(QtWidgets.QMessageBox):
    def __init__(self, parent, title, text):
        super(AboutBox, self).__init__(title, text, QtWidgets.QMessageBox.Information, 0, 0, 0, parent)

        if sys.platform.startswith('darwin'):
            self.setWindowFlags(QtCore.Qt.WindowTitleHint | QtCore.Qt.WindowSystemMenuHint)
        # Grab window icon of parent
        icon = self.windowIcon()
        size = icon.actualSize(QtCore.QSize(64, 64))
        self.setIconPixmap(icon.pixmap(size))

        # Add an invisible spacer to make the dialog box size to fit the width
        chars_per_line = 80
        fm = QtGui.QFontMetrics(self.font())
        leftMargin, topMargin, rightMargin, bottomMargin = self.getContentsMargins()
        width = fm.width('0') * chars_per_line + 4 + leftMargin + rightMargin

        horizontalSpacer = QtWidgets.QSpacerItem(width, 0, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        layout = self.layout()
        layout.addItem(horizontalSpacer, layout.rowCount(), 0, 1, layout.columnCount())

        self.setAttribute(QtCore.Qt.WA_DeleteOnClose)

class AboutBoxScrollbars(QtWidgets.QDialog):
    def __init__(self, parent, title, text, versiontext):
        super(AboutBoxScrollbars, self).__init__(parent)

        def _replace_urls(text):
            re_match_urls = re.compile(r"""((?:[a-z][\w-]+:(?:/{1,3}|[a-z0-9%])|www\d{0,3}[.]|[a-z0-9.\-]+[.‌​][a-z]{2,4}/)(?:[^\s()<>]+|(([^\s()<>]+|(([^\s()<>]+)))*))+(?:(([^\s()<>]+|(‌​([^\s()<>]+)))*)|[^\s`!()[]{};:'".,<>?«»“”‘’]))""", re.DOTALL)
            return re_match_urls.sub(lambda x: '<a href="%(url)s" style="color: #ffa02f;">%(url)s</a>' % dict(url=str(x.group())), text)

        if sys.platform.startswith('darwin'):
            self.setWindowFlags(QtCore.Qt.WindowTitleHint | QtCore.Qt.WindowSystemMenuHint)
        # Grab window icon of parent
        icon = self.windowIcon()
        size = icon.actualSize(QtCore.QSize(64, 64))

        self.setWindowTitle(title)

        label = QtWidgets.QLabel(self)
        label.setText(_replace_urls(text).replace('\n', '<br>'))
        label.setAlignment(QtCore.Qt.AlignVCenter | QtCore.Qt.AlignLeft)
        label.setOpenExternalLinks(True)
        label.setTextFormat(QtCore.Qt.RichText)

        if sys.platform.startswith('darwin'):
            label.setContentsMargins(16, 0, 0, 0)
        else:
            label.setContentsMargins(2, 0, 0, 0)
            label.setIndent(9)

        versionLabel = QtWidgets.QLabel(self)
        versionLabel.setText(versiontext)
        versionLabel.setContentsMargins(0, 0, 0, 10)
        f = versionLabel.font()
        f.setBold(True)
        versionLabel.setFont(f)

        iconLabel = QtWidgets.QLabel(self)
        iconLabel.setSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        iconLabel.setPixmap(icon.pixmap(size))
        iconLabel.setContentsMargins(5, 0, 5, 0)

        scroll = QtWidgets.QScrollArea(self)
        scroll.setWidget(label)
        scroll.setWidgetResizable(True)

        buttonBox = QtWidgets.QDialogButtonBox(QtWidgets.QDialogButtonBox.Ok)
        buttonBox.accepted.connect(self.close)
        buttonBox.button(QtWidgets.QDialogButtonBox.Ok).setDefault(True)
        buttonBox.setContentsMargins(0, 0, 10, 0)

        grid = QtWidgets.QGridLayout()
        if sys.platform.startswith('darwin'):
            grid.addWidget(versionLabel, 0, 1, 1, 1, QtCore.Qt.AlignTop)
            grid.addWidget(iconLabel, 0, 0, 2, 1, QtCore.Qt.AlignTop)
        else:
            grid.setContentsMargins(0, 0, 0, 0)
            grid.setVerticalSpacing(8)
            grid.setHorizontalSpacing(0)
            self.setContentsMargins(0, 15, 0, 20)
            grid.setRowStretch(1, 100)
            grid.setRowMinimumHeight(2, 6)

            grid.addWidget(versionLabel, 0, 1, 1, 1, QtCore.Qt.AlignTop | QtCore.Qt.AlignLeft)
            grid.addWidget(iconLabel, 0, 0, 2, 1, QtCore.Qt.AlignTop | QtCore.Qt.AlignLeft)

        grid.addWidget(scroll, 1, 1, 1, 1)
        grid.addWidget(buttonBox, 2, 0, 1, 2)

        if sys.platform.startswith('darwin'):
            f = self.font()
            f.setBold(True)
            label.setFont(f)

        # Add an invisible spacer to make the dialog box size to fit the width
        chars_per_line = 83
        fm = QtGui.QFontMetrics(label.font())
        leftMargin, topMargin, rightMargin, bottomMargin = self.getContentsMargins()
        width = fm.width('#') * chars_per_line + 4 + leftMargin + rightMargin
        horizontalSpacer = QtWidgets.QSpacerItem(width, 0, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        grid.addItem(horizontalSpacer, 3, 0, 1, 0)

        grid.setSizeConstraint(QtWidgets.QLayout.SetNoConstraint)
        self.setLayout(grid)

        self.setModal(True)
        self.setAttribute(QtCore.Qt.WA_DeleteOnClose)

        size = QtCore.QSize(min(parent.size().width(), width + 60), 0.8 * parent.size().height())
        self.resize(size)


class FileEntryView(QtWidgets.QWidget, Widget):
    """Widget for entering paths and filenames.

    It can be used in file save / load task views,
    for browsing directories and opening or saving files.
    """

    class FileSelectedEvent(events3d.Event):
        """Event class for the the File Entry that defines an event type
        to be emitted once the user confirms their choice on the file name."""

        def __init__(self, path, source):
            """Constructor of the FileSelected event.

            The event consists of the path of the selected file
            as well as the source that triggered the event,
            which can be 'browse' for when the browse button is clicked,
            'return' for when the return key is pressed while writing in
            the text edit, or 'button' for when the widget's button is
            clicked."""

            events3d.Event.__init__(self)
            self.path = pathToUnicode(path)
            self.source = source

        def __repr__(self):
            return "FileSelectedEvent: Path: %s Source: '%s'" % (
                repr(self.path), repr(self.source))

    def __init__(self, label='', buttonLabel='', mode='open' ):
        """FileEntryView constructor.

        The File Entry has a browse button on the left, which will open
        a standard save dialog. The center is occupied by an empty line edit,
        in which the user can simply enter a filename to save the file
        in the current directory. Finally, a button at the right can be used
        for accepting the filename and doing some action. You can control
        the button's caption with the buttonLabel parameter.

        By using the mode parameter, the widget will adapt to the needs,
        depending if it is used for opening a file, saving a file, or browsing
        a directory. Accordingly, its values can be: 'open', 'save', and 'dir'.

        If the mode is 'dir', there is no button. The final directory will be
        the path specified by the browse button.
        """

        super(FileEntryView, self).__init__()
        #Widget.__init__(self)

        # Declare data

        self._directory = ""
        self.label = TextView(label)

        # Define controls

        self.browse = BrowseButton()

        self.edit = QtWidgets.QLineEdit()
        self.edit.setValidator(
            QtGui.QRegExpValidator(QtCore.QRegExp(r'[^\/:*?"<>|]*'), None))

        buttonLabel = getLanguageString(buttonLabel)
        self.confirm = QtWidgets.QPushButton(buttonLabel)

        # Register events

        @self.browse.mhEvent
        def beforeBrowse(event):
            """
            Before the browse dialog is shown, make sure to set
            its path properly.
            """
            if self.mode == 'dir' or not self.text:
                self.browse.directory = self.directory
            else:
                self.browse.path = self.path

        @self.browse.mhEvent
        def onClicked(path):
            """
            When the browse button is used, update the path in
            the line edit and confirm the entry.
            """
            if path:
                self.path = pathToUnicode(path)
                self._confirm('browse')

        self.edit.returnPressed.connect(lambda: self._confirm('return'))

        self.edit.textEdited['QString'].connect(lambda s: self.callEvent('onChange', str(s)))

        self.confirm.clicked[bool].connect(lambda _: self._confirm('button'))

        # Create GUI

        self.setSizePolicy(QtWidgets.QSizePolicy.Ignored, QtWidgets.QSizePolicy.Fixed)

        self.layout = QtWidgets.QGridLayout(self)

        self.layout.addWidget(self.label, 0, 0)
        self.layout.setColumnStretch(0, 0)

        self.layout.addWidget(self.edit, 0, 1)
        self.layout.setColumnStretch(1, 1)

        self.layout.addWidget(self.browse, 0, 2)
        self.layout.setColumnStretch(2, 0)

        self.mode = mode

        if self.mode != 'dir':
            self.layout.addWidget(self.confirm, 0, 3)
            self.layout.setColumnStretch(3, 0)

    def getMode(self):
        """Get the FileEntryView's mode of operation."""
        return self.browse.mode

    def setMode(self, mode):
        """Set the FileEntryView's mode of operation."""
        self.browse.mode = mode

    mode = property(getMode, setMode)

    def getDirectory(self):
        """Get the FileEntryView's current directory."""
        return pathToUnicode(self._directory)

    def setDirectory(self, directory):
        """Set the directory that the widget will use for saving or
        opening the filename, and as an initial path for browsing in
        the dialog. If the mode is 'dir', the given path is written
        in the line edit."""
        self._directory = pathToUnicode(directory)
        if self.mode == 'dir':
            self.text = self.directory

    directory = property(getDirectory, setDirectory)

    def getPath(self):
        """Get the FileEntryView's selected path."""
        if self.mode == 'dir':
            return self.directory
        else:
            return os.path.normpath(os.path.join(self.directory, self.text))

    def setPath(self, path):
        """Set the path of the FileEntryView.
        This will update the widget's current directory,
        as well as the file name in the text edit."""
        path = pathToUnicode(path)
        if self.mode == 'dir':
            self.directory = path
        else:
            self.directory = os.path.dirname(path)
            self.text = os.path.basename(path)

    path = property(getPath, setPath)

    def getText(self):
        return str(self.edit.text())

    def setText(self, text):
        self.edit.setText(text)

    text = property(getText, setText)

    def getFilter(self):
        """Get the extension filter the widget uses for browsing."""
        return self.browse.filter

    def setFilter(self, filter):
        """Set the extension filter the widget will use for browsing."""
        self.browse.filter = filter

    filter = property(getFilter, setFilter)

    def _confirm(self, source):
        """Method to be called once the user has confirmed their choice,
        either by pressing the confirmation button, by pressing the return
        button in the line edit, or by using the browser.
        It emits an onFileSelected event if the path is not empty."""
        if self.mode == 'dir' and source in ('return', 'button'):
            self.directory = pathToUnicode(self.text)

        if len(self.text):
            self.callEvent('onFileSelected',
                self.FileSelectedEvent(self.path, source))
        else:
            log.notice("The text box is empty. Please enter a valid file name.")
            G.app.prompt("Missing Filename", "The text box is empty. Please enter a valid file name.", 'OK', helpId='noFileName')
            self.setFocus()

    def onFocus(self, event):
        """Handler for the event of the widget being given focus. It passes
        the focus directly to the line edit."""
        self.edit.setFocus()


class SplashScreen(QtWidgets.QSplashScreen):
    def __init__(self, image, version=""):
        super(SplashScreen, self).__init__(G.app.mainwin, getPixmap(image))
        self._stdout = sys.stdout
        self.messageRect = QtCore.QRect(118, 545, 620, 41)
        self.messageAlignment = QtCore.Qt.AlignLeft
        self.message = ''
        self.progressBarRect = QtCore.QRect(118, 576, 620, 8)
        self._version = version
        self.versionRect = QtCore.QRect(24, 547, 88, 36)
        self.setProgress(0)

    def setProgress(self, progress):
        self.progress = float(progress)
        if self.progress < 0:
            self.progress = 0.0
        if self.progress > 1:
            self.progress = 1.0

    def getProgress(self):
        return self.progress

    def escape(self, text):
        return text.replace('&','&amp;').replace('<','&lt;').replace('>','&gt;')

    def logMessage(self, text):
        text = self.escape(text)
        self.showMessage(text, alignment = QtCore.Qt.AlignHCenter)
        self.message = text

    def drawContents(self, painter):
        color = QtGui.QColor()
        color.setNamedColor('#ffffff')
        painter.setPen(color)
        font = painter.font()
        font.setPixelSize(16)
        painter.setFont(font)
        painter.drawText(self.messageRect, self.messageAlignment, self.message);
        if self._version:
            font.setPixelSize(29)
            painter.setFont(font)
            painter.drawText(self.versionRect, QtCore.Qt.AlignLeft, self._version)
        color.setNamedColor('#000000')
        painter.setPen(color)
        color.setNamedColor('#f58220')
        painter.setBrush(QtGui.QBrush(color))
        pRect = self.progressBarRect
        if self.getProgress() == 0:
            progressedWidth = 0
        else:
            progressedWidth = int(pRect.width()*self.getProgress())
        painter.drawRect(pRect.x(), pRect.y(), progressedWidth, pRect.height())
        color.setNamedColor('#191718')
        painter.setBrush(QtGui.QBrush(color))
        painter.drawRect(pRect.x()+progressedWidth, pRect.y(), pRect.width()-progressedWidth, pRect.height())

class StatusBar(QtWidgets.QStatusBar, Widget):
    def __init__(self):
        super(StatusBar, self).__init__()
        #Widget.__init__(self)
        self._perm = QtWidgets.QLabel()
        self.addWidget(self._perm, 1)
        self.duration = 2000
        self.__old_text = []

    def showMessage(self, text, *args):
        if isinstance(text,list):
            out = ""
            for part in text:
                out = out + getLanguageString(part)
            text = out
        else:
            text = getLanguageString(text)
        text = text % args
        super(StatusBar, self).showMessage(text, self.duration)

    def setMessage(self, text, *args):
        if isinstance(text,list):
            out = ""
            for part in text:
                out = out + getLanguageString(part)
            text = out
        else:
            text = getLanguageString(text)
        text = text % args
        self._perm.setText(text)

    def temporaryMessage(self, text, *args, msec=2000):
        timer = QtCore.QTimer()
        if isinstance(text,list):
            out = ""
            for part in text:
                out = out + getLanguageString(part)
            text = out
        else:
            text = getLanguageString(text)
        text = text % args
        self.__old_text.append(self._perm.text())
        self._perm.setText(text)
        timer.singleShot(msec, lambda: self._perm.setText(self.__old_text.pop()))

class VScrollLayout(QtWidgets.QLayout):
    def __init__(self, parent = None):
        super(VScrollLayout, self).__init__(parent)
        self.setContentsMargins(0, 0, 0, 0)
        self.setSizeConstraint(QtWidgets.QLayout.SetNoConstraint)
        self._child = None
        self._position = 0

    def addItem(self, item):
        if self._child is not None:
            raise RuntimeError('layout already has a child')
        self._child = item
        self._update()

    def count(self):
        return int(self._child is not None)

    def itemAt(self, index):
        if index != 0:
            return None
        if self._child is None:
            return None
        return self._child

    def takeAt(self, index):
        if self.child is None:
            return None
        child = self._child
        self._child = None
        self._update()
        return child

    def minimumSize(self):
        if self._child is None:
            return super(VScrollLayout, self).minimumSize()
        # log.debug('VScrollLayout.minimumSize(child): %d %d', self._child.sizeHint().width(), self._child.sizeHint().height())
        left, top, right, bottom = self.getContentsMargins()
        return QtCore.QSize(self._child.minimumSize().width() + left + right, 0)

    def maximumSize(self):
        if self._child is None:
            return super(VScrollLayout, self).maximumSize()
        # log.debug('VScrollLayout.maximumSize(child): %d %d', self._child.sizeHint().width(), self._child.sizeHint().height())
        left, top, right, bottom = self.getContentsMargins()
        return self._child.maximumSize() + QtCore.QSize(left + right, top + bottom)

    def sizeHint(self):
        if self._child is None:
            return super(VScrollLayout, self).sizeHint()
        # log.debug('VScrollLayout.sizeHint(child): %d %d', self._child.sizeHint().width(), self._child.sizeHint().height())
        left, top, right, bottom = self.getContentsMargins()
        return self._child.sizeHint() + QtCore.QSize(left + right, top + bottom)

    def setGeometry(self, rect):
        super(VScrollLayout, self).setGeometry(rect)
        # log.debug('VScrollLayout.setGeometry: position: %d', self._position)
        # log.debug('VScrollLayout.setGeometry: %d %d %d %d', rect.x(), rect.y(), rect.width(), rect.height())
        if self._child is None:
            return
        size = self._child.sizeHint()
        left, top, right, bottom = self.getContentsMargins()
        # log.debug('VScrollLayout.getContentsMargins: %d %d %d %d', left, top, right, bottom)

        rect = rect.adjusted(left, top, -right, -bottom)
        rect.adjust(0, -self._position, 0, -self._position)
        # log.debug("%x", int(self._child.widget().sizePolicy().horizontalPolicy()))
        if not self._child.widget().sizePolicy().horizontalPolicy() & QtWidgets.QSizePolicy.ExpandFlag:
            rect.setWidth(size.width())
        if not self._child.widget().sizePolicy().verticalPolicy() & QtWidgets.QSizePolicy.ExpandFlag:
            rect.setHeight(size.height())
        else:
            rect.setHeight(max(rect.height(), size.height()))

        # log.debug('VScrollLayout.setGeometry(child): %d %d %d %d', rect.x(), rect.y(), rect.width(), rect.height())
        self._child.setGeometry(rect)

    def expandingDirections(self):
        if self._child is None:
            return 0
        return self._child.expandingDirections()

    def hasHeightForWidth(self):
        if self._child is None:
            return super(VScrollLayout, self).hasHeightForWidth()
        return self._child.hasHeightForWidth()

    def heightForWidth(self, width):
        if self._child is None:
            return super(VScrollLayout, self).heightForWidth(width)
        return self._child.heightForWidth(width)

    def setPosition(self, value):
        self._position = value
        self._update()

    def _update(self):
        self.update()

    def childHeight(self):
        if self._child is None:
            return 0
        left, top, right, bottom = self.getContentsMargins()
        return self._child.sizeHint().height() + top + bottom

class Viewport(QtWidgets.QWidget):
    def __init__(self):
        super(Viewport, self).__init__()
        self.setContentsMargins(0, 0, 0, 0)
        self._layout = VScrollLayout(self)
        self._layout.setContentsMargins(1, 20, 1, 20)
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._child = None

    def setWidget(self, widget):
        if widget is None:
            self._layout.removeWidget(self._child)
        else:
            self._layout.addWidget(widget)
        self._child = widget
        self.updateGeometry()

    def childHeight(self):
        return self._layout.childHeight()

    def setPosition(self, value):
        self._layout.setPosition(value)

class VScrollArea(QtWidgets.QWidget, Widget):
    def __init__(self):
        super(VScrollArea, self).__init__()
        #Widget.__init__(self)

        self._viewport = Viewport()
        self._scrollbar = QtWidgets.QScrollBar(QtCore.Qt.Vertical)
        self._scrollbar.setRange(0, 0)
        self._scrollbar.setMinimumHeight(0)
        self._scrollbar.setSingleStep(10)
        self._layout = QtWidgets.QBoxLayout(QtWidgets.QBoxLayout.RightToLeft, self)
        self._layout.addWidget(self._scrollbar, 0)
        self._layout.addWidget(self._viewport, 1)
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._scrollbar.setTracking(True)
        self._widget = None
        self._scrollbar.valueChanged[int].connect(self._changed)

    def setWidget(self, widget):
        if self._widget is not None:
            self._widget.removeEventFilter(self)
        self._widget = widget
        self._viewport.setWidget(self._widget)
        self.updateGeometry()
        self._updateScrollSize()
        if self._widget is not None:
            self._widget.installEventFilter(self)

    def resizeEvent(self, event):
        # log.debug('resizeEvent: %d, %d', event.size().width(), event.size().height())
        super(VScrollArea, self).resizeEvent(event)
        self._updateScrollSize()
        self._updateScrollPosition()

    def _updateScrollSize(self):
        cheight = self._viewport.childHeight()
        vheight = self._viewport.size().height()
        # log.debug('_updateScrollSize: %d, %d', cheight, vheight)
        self._scrollbar.setRange(0, cheight - vheight)
        self._scrollbar.setPageStep(vheight)

    def _changed(self, value):
        # log.debug('VScrollArea_changed: %d', value)
        self._updateScrollPosition()

    def _updateScrollPosition(self):
        value = self._scrollbar.value()
        # log.debug('_updateScrollPosition: %d', value)
        self._viewport.setPosition(value)

    def eventFilter(self, object, event):
        if object == self._widget and event.type() != QtCore.QEvent.Resize:
            # log.debug('Viewport child resize: %d,%d -> %d,%d',
            #           event.oldSize().width(), event.oldSize().height(),
            #           event.size().width(), event.size().height())
            self._updateScrollSize()
        return False

    def getClassName(self):
        """
        Classname for this widet, useful for styling using qss.
        """
        return str(self.metaObject().className())

class TreeItem(QtWidgets.QTreeWidgetItem):
    def __init__(self, text, parent=None, isDir=False):
        super(TreeItem, self).__init__([text])
        self.text = text
        self.parent = parent
        self.isDir = isDir
        if self.isDir:
            self.setIcon(0, TreeView._dirIcon)
            self.setChildIndicatorPolicy(QtWidgets.QTreeWidgetItem.ShowIndicator)
        else:
            self.setIcon(0, TreeView._fileIcon)
            self.setChildIndicatorPolicy(QtWidgets.QTreeWidgetItem.DontShowIndicator)

    def addChild(self, text, isDir=False):
        item = TreeItem(text, self, isDir)
        super(TreeItem, self).addChild(item)
        return item

    def addChildren(self, strings):
        items = [TreeItem(text, self) for text in strings]
        super(TreeItem, self).addChildren(items)
        return items

class TreeView(QtWidgets.QTreeWidget, Widget):
    _dirIcon = None
    _fileIcon = None

    def __init__(self, parent = None):
        super(TreeView, self).__init__(parent)
        #Widget.__init__(self)
        self.itemActivated[QtWidgets.QTreeWidgetItem, int].connect(self._activate)
        self.itemExpanded[QtWidgets.QTreeWidgetItem].connect(self._expand)
        if TreeView._dirIcon is None:
            TreeView._dirIcon = self.style().standardIcon(QtWidgets.QStyle.SP_DirIcon)
        if TreeView._fileIcon is None:
            TreeView._fileIcon = self.style().standardIcon(QtWidgets.QStyle.SP_FileIcon)

    def addTopLevel(self, text, isDir=True):
        item = TreeItem(text, None, isDir)
        self.addTopLevelItem(item)
        return item

    def _activate(self, item, column):
        if not item.isDir:
            self.callEvent('onActivate', item)

    def _expand(self, item):
        if item.isDir:
            self.callEvent('onExpand', item)

class SpinBox(QtWidgets.QSpinBox, Widget):
    def __init__(self, value, parent = None):
        super(SpinBox, self).__init__(parent)
        #Widget.__init__(self)
        self.setRange(0, 99999)
        self.setValue(value)
        self.valueChanged[int].connect(self._changed)

    def _changed(self, value):
        self.callEvent('onChange', value)

    def setValue(self, value):
        self.blockSignals(True)
        super(SpinBox, self).setValue(value)
        self.blockSignals(False)

class BrowseButton(Button):

    @staticmethod
    def conformFilter(filter):
        """Static method that corrects a filter string, if needed, by
        translating it and adding the default 'All Files' filter to it."""
        filter = getLanguageString(filter)
        if '(*.*)' not in filter:
            filter = ';;'.join([filter, getLanguageString('All Files') + ' (*.*)'])
        return filter

    @staticmethod
    def getExistingPath(path):
        path = pathToUnicode(path)
        if not os.path.isdir(path) and not os.path.isfile(path):
            path = os.path.split(path)[0]
            homePath = os.path.abspath(getPath(''))
            if os.path.isdir(homePath) and isSubPath(os.path.abspath(path), homePath):
                # Find first existing folder within MH home path
                while path and not os.path.isdir(path):
                    path = os.path.split(path)[0]
            if not os.path.isdir(path):
                path = os.getcwd()
        return os.path.normpath(path)

    def __init__(self, mode = 'open', label=None, returnFilter=False):
        if label is None:
            label = '...'
        super(BrowseButton, self).__init__(label)
        self._directory = self.getExistingPath("")
        self.filename = ""
        self._filter = ''
        self._mode = None
        self._returnFilter = returnFilter
        self.mode = mode

    def getMode(self):
        return self._mode

    def setMode(self, mode):
        mode = mode.lower()
        if mode not in ('open', 'save', 'dir'):
            raise RuntimeError("mode '%s' not recognised; must be 'open', 'save', or 'dir'")
        self._mode = mode

    mode = property(getMode, setMode)

    def getDirectory(self):
        """Get the directory in which the dialog will browse."""
        return self._directory

    def setDirectory(self, dir):
        """Set the directory in which the dialog will browse."""
        self._directory = pathToUnicode(dir)

    directory = property(getDirectory, setDirectory)

    def getPath(self):
        return os.path.normpath(os.path.join(self.directory, self.filename))

    def setPath(self, path):
        """WARNING: Use only with complete file paths that include filename."""
        self.directory = os.path.dirname(pathToUnicode(path))
        self.filename = os.path.basename(pathToUnicode(path))

    path = property(getPath, setPath)

    def getFilter(self):
        return self._filter

    def setFilter(self, filter):
        self._filter = self.conformFilter(filter)

    filter = property(getFilter, setFilter)

    def retrunFilter(self, b=None):
        if b is None:
            return self._returnFilter
        elif isinstance(b, bool):
            self._returnFilter = b

    def _clicked(self, state):
        self.callEvent('beforeBrowse', None)

        path = self.getExistingPath(self.path)
        if self.mode == 'save' and self.filename:
            # Don't discard filename when saving
            path = os.path.join(path, self.filename)

        if self.mode == 'open':
            path = QtWidgets.QFileDialog.getOpenFileName(G.app.mainwin, directory=self.directory, filter=self.filter)
        elif self.mode == 'save':
            path = QtWidgets.QFileDialog.getSaveFileName(G.app.mainwin, directory=self.directory, filter=self.filter)
        elif self.mode == 'dir':
            path = QtWidgets.QFileDialog.getExistingDirectory(G.app.mainwin, directory=self.directory)

        if path:
            if self._returnFilter and self.mode != 'dir':
                self.callEvent('onClicked', path)
            else:
                if self.mode == 'dir':
                    self.directory = path
                    self.callEvent('onClicked', path)
                else:
                    self.callEvent('onClicked', path[0])


class ColorPickButton(Button):
    """
    Button widget that opens a color picker when clicked.
    """
    # TODO add a rectangle showing the current color

    def __init__(self, initialColor = None):
        super(ColorPickButton, self).__init__("Pick")
        if initialColor is not None:
            self.color = initialColor
        else:
            import material
            self.color = material.Color()

    def getColor(self):
        return self._color

    def setColor(self, color):
        import material
        if isinstance(color, material.Color):
            self._color = color
        elif isinstance(color, QtGui.QColor):
            self._color = colorFromQColor(color)
        else:
            self._color = material.Color().copyFrom(color)

    color = property(getColor, setColor)

    def _clicked(self, state):
        currentColor = qColorFromColor(self.color)
        pickedColor = QtWidgets.QColorDialog.getColor(currentColor)
        if pickedColor.isValid():
            self.color = colorFromQColor(pickedColor)
            self.callEvent('onClicked', self.color)

class Action(QtWidgets.QAction, Widget):
    _groups = {}

    @classmethod
    def getIcon(cls, name):
        from qtui import supportsSVG
        # icon = G.app.mainwin.style().standardIcon(QtGui.QStyle.SP_MessageBoxWarning)
        svgPath = os.path.join(getSysDataPath('icons'), name + '.svg')
        if supportsSVG() and os.path.isfile(svgPath):
            path = svgPath
        else:
            path = os.path.join(getSysDataPath('icons'), name + '.png')

        if G.app.theme:
            themePath = os.path.join(getSysDataPath('themes'), G.app.theme, 'icons', name + '.svg')
            if supportsSVG() and os.path.isfile(themePath):
                path = themePath
            else:
                themePath = os.path.join(getSysDataPath('themes'), G.app.theme, 'icons', name + '.png')
                if os.path.isfile(themePath):
                    path = themePath

        if not os.path.isfile(path):
            path = os.path.join(getSysDataPath('icons'), 'notfound.png')
        icon = QtGui.QIcon(path)

        # Allows setting custom icons for active, selected and disabled states
        ext = os.path.splitext(path)[1]
        for (name, mode) in [ ("disabled", QtGui.QIcon.Disabled),
                              ("active", QtGui.QIcon.Active),
                              ("selected", QtGui.QIcon.Selected) ]:
            customIconPath = "%s_%s%s" % (os.path.splitext(path)[0], name, ext)
            if os.path.isfile(customIconPath):
                icon.addFile(customIconPath, QtCore.QSize(), mode)
        return icon

    @classmethod
    def getGroup(cls, name):
        if name not in cls._groups:
            cls._groups[name] = ActionGroup()
        return cls._groups[name]

    def __init__(self, name, text, method, tooltip = None, group = None, toggle = False):
        super(Action, self).__init__(self.getIcon(name), text, G.app.mainwin)
        self.name = name
        self.method = method
        if tooltip is not None:
            self.setToolTip(tooltip)
        if group is not None:
            self.setActionGroup(self.getGroup(group))
        if toggle:
            self.setCheckable(True)
        self.triggered[bool].connect(self._activate)

    @property
    def text(self):
        return str(super(Action, self).text())

    def setActionGroup(self, group):
        self.setCheckable(True)
        super(Action, self).setActionGroup(group)

    def _activate(self, checked):
        self.method()

class ActionGroup(QtWidgets.QActionGroup):
    def __init__(self):
        super(ActionGroup, self).__init__(G.app.mainwin)

class Actions(object):
    def __init__(self):
        self._order = []

    def __setattr__(self, name, value):
        if name[0] != '_':
            self._order.append(value)
        object.__setattr__(self, name, value)

    def __iter__(self):
        return iter(self._order)

class SizePolicy(object):
    Fixed               = QtWidgets.QSizePolicy.Fixed
    Minimum             = QtWidgets.QSizePolicy.Minimum
    Maximum             = QtWidgets.QSizePolicy.Maximum
    Preferred           = QtWidgets.QSizePolicy.Preferred
    Expanding           = QtWidgets.QSizePolicy.Expanding
    MinimumExpanding    = QtWidgets.QSizePolicy.MinimumExpanding
    Ignored             = QtWidgets.QSizePolicy.Ignored

class TableItem(QtWidgets.QTableWidgetItem):
    def setUserData(self, data):
        self.setData(QtCore.Qt.UserRole, data)

    def getUserData(self):
        return self.data(QtCore.Qt.UserRole)

    @property
    def text(self):
        return str(super(TableItem, self).text())

class TableView(QtWidgets.QTableWidget, Widget):
    def __init__(self):
        super(TableView, self).__init__()
        #Widget.__init__(self)

    def setItem(self, row, col, text, data = None):
        item = TableItem(text)
        if data is not None:
            item.setUserData(data)
        super(TableView, self).setItem(row, col, item)

    def getItemData(self, row, col):
        return self.item(row, col).getUserData()

class ImageView(QtWidgets.QLabel, QtWidgets.QScrollArea, Widget):
    def __init__(self):
        super(ImageView, self).__init__()
        #Widget.__init__(self)
        self.setAlignment(QtCore.Qt.AlignHCenter | QtCore.Qt.AlignTop)
        self.setMinimumSize(50,50)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Ignored, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHeightForWidth(True)
        self.setSizePolicy(sizePolicy)
        #self.setScaledContents(True)
        self.ratio = 1.0
        self.workingSize = None
        self._pixmap = None

    def setImage(self, img):
        self._pixmap = getPixmap(img)
        self.adjustSize()
        self.updateGeometry()
        self.refreshImage()

    def sizeHint(self):
        if not self._pixmap:
            return super(ImageView, self).sizeHint()
        return self._pixmap.size()

    def heightForWidth(self, width):
        if not self._pixmap or self._pixmap.width() == 0:
            return width
        else:
            size = self._pixmap.size()
            return int((float(width)/size.width())*size.height())

    def resizeEvent(self, event):
        self.workingSize = event.size()
        self.refreshImage(event.size())

    def refreshImage(self, size = None):
        if not self._pixmap:
            return

        if not size:
            size = self.sizeHint()

        pixmap = self._pixmap
        w = pixmap.width()
        h = pixmap.height()
        size *= self.ratio
        if w > size.width() or h > size.height():
            pixmap = pixmap.scaled(size.width(), size.height(), QtCore.Qt.KeepAspectRatio)
        self.setPixmap(pixmap)

    def save(self, fname):
        if not os.path.splitext(fname)[1]:
            fname = fname + '.png'

        if self._pixmap:
            self._pixmap.save (fname)


class ZoomableImageView(QtWidgets.QScrollArea, Widget):
    def __init__(self):
        QtWidgets.QScrollArea.__init__(self)
        #Widget.__init__(self)
        self.imageLabel = QtWidgets.QLabel()
        self.imageLabel.setBackgroundRole(QtGui.QPalette.Base)
        self.imageLabel.setAlignment(QtCore.Qt.AlignHCenter | QtCore.Qt.AlignTop)
        self.imageLabel.setMinimumSize(5,5)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Ignored, QtWidgets.QSizePolicy.Ignored)
        sizePolicy.setHeightForWidth(True)
        self.imageLabel.setSizePolicy(sizePolicy)
        self.imageLabel.setScaledContents(True)
        self.setBackgroundRole(QtGui.QPalette.Dark)
        self.setWidget(self.imageLabel)
        self.workingWidth = None
        self.ratio = 1.0
        self.minratio = 0.1

    def setImage(self, img):
        pixmap = getPixmap(img)
        self.imageLabel.setPixmap(pixmap)
        self.imageLabel.adjustSize()
        self.ratio = float(self.width()) / float(pixmap.width())
        self.minratio = 100.0 / float(pixmap.width())
        if self.minratio > 1.0:
            self.minratio = 1.0
        self.imageLabel.updateGeometry()
        self.refreshImage()

    def resizeEvent(self, event):
        if self.workingWidth:
            self.ratio *= float(event.size().width()) / float(self.workingWidth)
        self.workingWidth = event.size().width()
        if self.ratio > 1.0:
            self.ratio = 1.0
        self.refreshImage(True)

    def heightForWidth(self, width):
        pixmap = self.imageLabel.pixmap()
        if not pixmap or pixmap.width() == 0:
            return width
        else:
            size = pixmap.size()
            return int((float(width)/size.width())*size.height())

    def refreshImage(self, zoomAct = False, displ = (0, 0)):
        pixmap = self.imageLabel.pixmap()
        if not pixmap:
            return
        scrat = [0.0, 0.0]
        if zoomAct:
            for (index, scrollBar) in enumerate((self.horizontalScrollBar(), self.verticalScrollBar())):
                if scrollBar.maximum() > 0:
                    scrat[index] = float(scrollBar.value()) / float(scrollBar.maximum())
                else:
                    scrat[index] = 0.5
        self.imageLabel.resize(self.ratio * pixmap.size())
        if zoomAct:
            for (index, scrollBar) in enumerate((self.horizontalScrollBar(), self.verticalScrollBar())):
                if scrollBar.maximum() > 0:
                    scrollBar.setValue(int(scrat[index] * scrollBar.maximum() + displ[index]))

    def save(self, fname, iformat=None):
        if not os.path.splitext(fname)[1]:
            fname = fname + '.png'
        if self.imageLabel.pixmap():
            self.imageLabel.pixmap().save(fname, format=iformat)

    def mousePressEvent(self, event):
        self.mdown = event.pos()

    def mouseMoveEvent(self, event):
        if event.buttons() & QtCore.Qt.RightButton:
            self.wheelEvent(QtGui.QWheelEvent(event.pos(),
                                              10 * (event.pos().y() - self.mdown.y()),
                                              event.buttons(), event.modifiers()), False)
            self.mdown = event.pos()
        else:
            self.horizontalScrollBar().setValue(
                self.horizontalScrollBar().value() + 2*(self.mdown.x() - event.pos().x()))
            self.verticalScrollBar().setValue(
                self.verticalScrollBar().value() + 2*(self.mdown.y() - event.pos().y()))
            self.mdown = event.pos()
            self.refreshImage()

    def wheelEvent(self, event, displace = True):
        ratbef = self.ratio
        if G.app.getSetting('invertMouseWheel'):
            delta = event.angleDelta()
        else:
            delta = -event.angleDelta()
        factor = 1 - delta.y()*0.0007
        self.ratio *= factor
        if self.ratio > 1.0:
            self.ratio = 1.0
        if self.ratio < self.minratio:
            self.ratio = self.minratio
        dr = 2*abs(self.ratio - ratbef) if displace else 0
        self.refreshImage(True,
                          (dr*(event.x() - self.width()/2),
                           dr*(event.y() - self.height()/2)))


def colorFromQColor(qColor):
    import material
    if qColor.isValid():
        values = (float(qColor.red())/255,
                  float(qColor.green())/255,
                  float(qColor.blue())/255)
        return material.Color().copyFrom(values)
    else:
        return material.Color()

def qColorFromColor(color):
    import material
    if isinstance(color, material.Color):
        color = color.asTuple()
    return QtGui.QColor(int(color[0]*255),
                            int(color[1]*255),
                            int(color[2]*255))

def getPixmap(img):
    import image
    if isinstance(img, image.Image):
        img = img.toQImage()
        return QtGui.QPixmap.fromImage(img)
    elif isinstance(img, QtGui.QPixmap):
        return img
    elif isinstance(img, QtGui.QImage):
        return QtGui.QPixmap.fromImage(img)
    else:
        return QtGui.QPixmap(pathToUnicode(img))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
**Project Name:**      MakeHuman

**Product Home Page:** http://www.makehumancommunity.org/

**Github Code Home Page:**    https://github.com/makehumancommunity/

**Authors:**           Glynn Clements

**Copyright(c):**      MakeHuman Team 2001-2020

**Licensing:**         AGPL3

    This file is part of MakeHuman Community (www.makehumancommunity.org).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.


Abstract
--------

TODO
"""

import sys
sys.path = [".", "./core", "./lib"] + sys.path
import makehuman
import algos3d
import numpy as np
import os
import zipfile
import fnmatch

def getAllFiles(rootPath, filterStrArr):
    result = [ None ]*len(filterStrArr)
    for root, dirnames, filenames in os.walk(rootPath):
        for i, filterStr in enumerate(filterStrArr):
            if not result[i]:
                result[i] = []
            result[i].extend(getFiles(root, filenames, filterStr))
    return result

def getFiles(root, filenames, filterStr):
    foundFiles = []
    for filename in fnmatch.filter(filenames, filterStr):
        foundFiles.append(os.path.join(root, filename))
    return foundFiles


if __name__ == '__main__':
    obj = algos3d.Target(None, None)
    allFiles = getAllFiles('data', ['*.target', '*.png'])
    npzPath = 'data/targets.npz'
    with zipfile.ZipFile(npzPath, mode='w', compression=zipfile.ZIP_DEFLATED) as zip:
        npzdir = os.path.dirname(npzPath)
        allTargets = allFiles[0]

        # License for all official MH targets
        lpath = 'data/targets/targets.license.npy'
        np.save(lpath, makehuman.getAssetLicense().toNumpyString())
        zip.write(lpath, os.path.relpath(lpath, npzdir))
        os.remove(lpath)

        for (i, path) in enumerate(allTargets):
            try:
                obj._load_text(path)
                iname, vname, lname = obj._save_binary(path)
                zip.write(iname, os.path.relpath(iname, npzdir))
                zip.write(vname, os.path.relpath(vname, npzdir))
                if lname:
                    zip.write(lname, os.path.relpath(lname, npzdir))
                    os.remove(lname)
                os.remove(iname)
                os.remove(vname)
                print("[%.0f%% done] converted target %s" % (100*(float(i)/float(len(allTargets))), path))
            except None as e:
                raise e
                print('error converting target %s' % path)

    print("Writing images list")
    with open('data/images.list', 'w', encoding="utf-8") as f:
        allImages = allFiles[1]
        for path in allImages:
            path = path.replace('\\','/')
            f.write(path + '\n')
    print("All done.")
